import functools
import os
import hashlib
import hmac
import math
import struct
from functools import reduce
from pyasn1.codec.der import decoder, encoder
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from Crypto.Cipher import AES
from Crypto.Hash import HMAC, SHA1, MD5
from .rc4 import RC4
from .gss import gss_wrap_rc4, gss_unwrap_rc4
from ..exceptions import KerberosErr


def _nfold(ba, nbytes):
    def rotate_right(ba, nbits):
        ba = bytearray(ba)
        n = len(ba)
        nbits %= (n * 8)
        # Rotate bytes
        for _ in range(nbits // 8):
            ba = bytearray([ba[-1]]) + ba[:-1]
        # Rotate remaining bits
        nbits %= 8
        if nbits > 0:
            last_bit = 0
            for i in range(len(ba)):
                new_last_bit = ba[i] & ((1 << nbits) - 1)
                ba[i] = (ba[i] >> nbits) | (last_bit << (8 - nbits))
                last_bit = new_last_bit
            ba[0] |= (last_bit << (8 - nbits))
        return ba

    def add_ones_complement(ba1, ba2):
        n = len(ba1)
        res = [0] * n
        carry = 0
        for i in range(n-1, -1, -1):
            s = ba1[i] + ba2[i] + carry
            res[i] = s & 0xff
            carry = s >> 8

        pos = n - 1
        while carry and pos >= 0:
            s = res[pos] + carry
            res[pos] = s & 0xff
            carry = s >> 8
            pos -= 1
        return bytearray(res)

    slen = len(ba)

    def lcm(a, b):
        return abs(a*b) // math.gcd(a, b)

    lcm_val = lcm(slen, nbytes)
    big_str = bytearray()
    curr = bytearray(ba)
    for i in range(lcm_val // slen):
        big_str += curr
        curr = rotate_right(curr, 13)

    parts = [big_str[i:i+nbytes] for i in range(0, len(big_str), nbytes)]
    return bytes(reduce(add_ones_complement, parts))


def derive_key(base_key, usage_int, payload_byte=None):
    constant = struct.pack('>I', usage_int)
    if payload_byte is not None:
        constant += bytes([payload_byte])

    nfolded = _nfold(constant, 16)
    cipher = Cipher(algorithms.AES(base_key), modes.ECB())
    encryptor1 = cipher.encryptor()
    b1 = encryptor1.update(nfolded) + encryptor1.finalize()
    encryptor2 = cipher.encryptor()
    b2 = encryptor2.update(b1) + encryptor2.finalize()

    return b1 + b2


def aes_cts_encrypt(key: bytes, plain: bytes) -> bytes:
    n = len(plain)  # 43 bytes
    iv = b'\x00' * 16
    pad_len = 16 - (n % 16)
    padded_plain = plain + b'\x00' * pad_len

    # CBC encryptie (correct IV)
    cipher = Cipher(algorithms.AES(key), modes.CBC(iv))
    encryptor = cipher.encryptor()
    full_cipher = encryptor.update(padded_plain) + encryptor.finalize()

    c1 = full_cipher[0:16]
    c2 = full_cipher[16:32]
    c3 = full_cipher[32:48]

    return c1 + c3 + c2[:n % 16]


def decrypt_kerberos_aes_cts(key: bytes, usage: int, cipher: bytes) -> bytes:
    ciphertext = cipher[:-12]
    expected_hmac = cipher[-12:]

    ki = derive_key(key, usage, 0x55)
    ke = derive_key(key, usage, 0xAA)

    n = len(ciphertext)
    block_size = 16

    aes = AES.new(ke, AES.MODE_ECB)

    if n % block_size == 0:
        aes_cbc = AES.new(ke, AES.MODE_CBC, b'\x00' * block_size)
        plaintext_with_confounder = aes_cbc.decrypt(ciphertext)
    else:
        m = n // block_size
        last_blocks_len = n % block_size

        iv = b'\x00' * block_size
        plaintext_with_confounder = b''

        for i in range(m - 1):
            block = ciphertext[i*block_size: (i+1)*block_size]
            dec = aes.decrypt(block)
            plaintext_with_confounder += \
                bytes(a ^ b for a, b in zip(dec, iv))
            iv = block

        cn_minus_1 = ciphertext[(m-1)*block_size: m*block_size]
        cn = ciphertext[m*block_size:]

        dec_intermediate = aes.decrypt(cn_minus_1)

        pn = bytes(
            a ^ b for a, b in zip(dec_intermediate[:last_blocks_len], cn))

        stolen_bytes = dec_intermediate[last_blocks_len:]
        full_block_to_decrypt = cn + stolen_bytes

        dec_pn_minus_1 = aes.decrypt(full_block_to_decrypt)
        pn_minus_1 = bytes(a ^ b for a, b in zip(dec_pn_minus_1, iv))

        plaintext_with_confounder += pn_minus_1 + pn

    h = HMAC.new(ki, plaintext_with_confounder, SHA1)
    if h.digest()[:12] != expected_hmac:
        raise ValueError("Integrity check failed: HMAC mismatch")

    return plaintext_with_confounder


def encrypt_kerberos_aes_cts(session_key: bytes,
                             usage: int,
                             plain_text: bytes):
    ki = derive_key(session_key, usage, 0x55)
    ke = derive_key(session_key, usage, 0xAA)

    confounder = os.urandom(16)
    basic_plaintext = confounder + plain_text

    h = HMAC.new(ki, basic_plaintext, SHA1)
    checksum = h.digest()[:12]

    aes = AES.new(ke, AES.MODE_CBC, b'\x00' * 16)
    n = len(basic_plaintext)

    if n % 16 == 0:
        final_ctext = aes.encrypt(basic_plaintext)
    else:
        pad_len = 16 - (n % 16)
        padded_data = basic_plaintext + b'\x00' * pad_len
        ctext = aes.encrypt(padded_data)

        last_full_block = ctext[-32:-16]
        truncated_block = ctext[-16:]

        final_ctext = (
            ctext[:-32] +
            truncated_block +
            last_full_block[:16-pad_len]
        )

    return final_ctext + checksum


def encrypt_kerberos_rc4(session_key, usage, plaintext):
    confounder = os.urandom(8)
    data_to_encrypt = confounder + plaintext

    usage_bytes = struct.pack('<I', usage)
    k1 = HMAC.new(session_key, usage_bytes, MD5).digest()

    checksum = HMAC.new(k1, data_to_encrypt, MD5).digest()
    k3 = HMAC.new(k1, checksum, MD5).digest()

    cipher = RC4(k3)
    encrypted_data = cipher.encrypt(data_to_encrypt)

    return checksum + encrypted_data


def decrypt_kerberos_rc4(key, usage, data):
    checksum = data[:16]
    encrypted_data = data[16:]

    k1 = hmac.new(key, struct.pack('<I', usage), hashlib.md5).digest()
    k3 = hmac.new(k1, checksum, hashlib.md5).digest()

    cipher = RC4(k3)
    decrypted = cipher.decrypt(encrypted_data)

    return decrypted[8:]  # Skip 8 bytes confounder


def read_session_key(data: bytes) -> bytes:
    payload = data[16:]
    as_rep_part, _ = decoder.decode(payload)
    key_container_bytes = encoder.encode(as_rep_part[0])
    inner_key_seq, _ = decoder.decode(key_container_bytes)

    key_type_bytes = encoder.encode(inner_key_seq[0])
    key_value_bytes = encoder.encode(inner_key_seq[1])

    key_type_obj, _ = decoder.decode(key_type_bytes)
    key_value_obj, _ = decoder.decode(key_value_bytes)

    session_key = bytes(key_value_obj.asOctets())
    _key_type = int(key_type_obj)

    return session_key


def seal_func_kerberos(session_key: bytes):
    def _kerberos_sealer(
            flags: int,
            seq_num: int,
            message_to_sign: bytes,
            message_to_encrypt: bytes,
            session_key: bytes) -> tuple[bytes, bytes]:
        return gss_wrap_rc4(session_key,
                            message_to_encrypt,
                            seq_num)
    return functools.partial(_kerberos_sealer, session_key=session_key)


def gss_unwrap_kerberos(session_key: bytes):
    def _kerberos_unwrap(
            cipher_text: bytes,
            auth_data: bytes,
            session_key: bytes) -> tuple[bytes, bytes]:
        return gss_unwrap_rc4(session_key,
                              cipher_text,
                              auth_data)
    return functools.partial(_kerberos_unwrap, session_key=session_key)


def sign_func_kerberos(session_key: bytes):
    def _kerberos_signer(
            flags: int,
            seq_num: int,
            message_to_sign: bytes,
            session_key: bytes) -> tuple[bytes, bytes]:
        return gss_wrap_rc4(session_key,
                            message_to_sign,
                            seq_num)
    return functools.partial(_kerberos_signer, session_key=session_key)


KDC_ERR_PREAUTH_REQUIRED = 25
KRB_ERRORS = {
    6:  "KDC_ERR_C_PRINCIPAL_UNKNOWN (Username not found)",
    7:  "KDC_ERR_S_PRINCIPAL_UNKNOWN (Service not found; Must use FQDN !!)",
    18: "KDC_ERR_CLIENT_REVOKED (Account blocked)",
    24: "KDC_ERR_PREAUTH_FAILED (Wrong password)",
    31: "KDC_ERR_S_PRINCIPAL_UNKNOWN (Service SPN not found)",
    60: "KDC_ERR_WRONG_REALM (Wrong domain name)"
}


def parse_krb_error(data: bytes):
    if data[0] != 0x7e:
        return

    code = -1
    pos = 2 if data[1] < 128 else 2 + (data[1] & 0x7f)
    pos += 2 if data[pos+1] < 128 else 2 + (data[pos+1] & 0x7f)

    while pos < len(data):
        tag = data[pos]
        length = data[pos+1]
        if tag == 0xa6:
            code = data[pos + 4]
            if code == KDC_ERR_PREAUTH_REQUIRED:
                return
            break
        pos += 2 + length

    raise KerberosErr(
        KRB_ERRORS.get(code, f"Unknown Kerberos Error: {code}")
    )



from __future__ import annotations
import struct
from typing import TYPE_CHECKING
from .common import RpcCommon
from .baseresp import RpcBaseResp
from ..exceptions import BindNak


if TYPE_CHECKING:
    from ..dcom import Dcom


class RpcBindNak(RpcBaseResp):

    def __init__(self, dcom: Dcom, rpc_common: RpcCommon, data: bytes):
        if len(data) < 24:
            raise BindNak("Packet too short")

        header = struct.unpack('<BBBBIHHI', data[:16])
        pdu_type = header[2]
        call_id = header[7]

        if pdu_type != 13:
            raise BindNak(
                f"Error: PDU type is {pdu_type}, expecting 13 (Bind_nak).")

        provider_reason_code = struct.unpack('<H', data[16:18])[0]
        status_code = struct.unpack('<H', data[18:20])[0]

        provider_reasons = {
            0: "REASON_NOT_SPECIFIED (General error)",
            1: "TEMPORARY_CONGESTION (Server to0 busy)",
            2: "LOCAL_LIMIT_EXCEEDED (Too many sessions)",
            3: "PROTOCOL_VERSION_NOT_SUPPORTED (Protocol version mismatch)",
            4: "AUTHENTICATION_TYPE_NOT_SUPPORTED (Authentication denied)",
            5: "INVALID_AUTH_INSTANCE (Invalid authentication data)"
        }

        windows_errors = {
            0x0501: "RPC_S_PROTSEQ_NOT_SUPPORTED (Security flags denied)",
            0x0005: "ERROR_ACCESS_DENIED (Access denied)",
            0x06d3:
            "RPC_S_AUTHN_LEVEL_LOW (Auth level too low, PKT_PRIVACY required)",
            0x1c01000b:
            "RPC_S_WRONG_KIND_OF_AUTH (Wrong authentication method)"
        }

        reason_text = provider_reasons.get(provider_reason_code,
                                           f"Unknown ({provider_reason_code})")
        error_text = windows_errors.get(status_code,
                                        "Unknown Windows error")

        output = [
            "--- MSRPC BIND_NAK ---",
            f"Call ID         : {call_id}",
            f"Provider Reason : {reason_text}",
            f"Windows Status  : 0x{status_code:04x} -> {error_text}",
            "------------------------------"
        ]

        raise BindNak("\n".join(output))



import struct
from typing import OrderedDict
from .orpcthat import ORPCTHAT
from .objref_standard import ObjRefStandard
from .activation_blob import ActivationBlob
from .scm_reply_info_data import ScmReplyInfoData
from .props_out_info import PropsOutInfo
from ..tools import is_fqdn
from .varying_array import VaryingArray
from .encoding_unit import EncodingUnit
from .property_info import PropertyInfo
from .next_response import NextResponse
from .object_block import ObjectBlock


class NextBigResponse(NextResponse):

    _FMT1_32 = '<LLL'
    _FMT1_64 = '<QQQ'

    _FMT1_32_SZ = struct.calcsize(_FMT1_32)
    _FMT1_64_SZ = struct.calcsize(_FMT1_64)

    def __init__(self, data: bytes):
        self.orpcthat, offset = ORPCTHAT.from_data(data, offset=0)
        (
            ncount,
            noffset,
            ncount,
        ) = struct.unpack_from(self._FMT1_32, data, offset)
        offset += self._FMT1_32_SZ
        ap_objects = []

        for _ in range(ncount):
            va, offset = VaryingArray.from_data(data, offset)
            encoding_unit = EncodingUnit(va.objref.object_data)
            ap_objects.append(encoding_unit)

        (
            self.pu_returned,
            self.error_code,
        ) = struct.unpack_from('<LL', data, offset)

        assert len(ap_objects) == 1, "only support for a single object block"
        self._obj_block = ap_objects[0].object_block

    def _get_object_block(self) -> ObjectBlock:
        return self._obj_block



import struct
from collections import OrderedDict
from typing import Tuple
from ..cim_type import CimType
from ..const import DICTIONARY_REFERENCE
from .encoded_string import EncodedString
from .property_info import PropertyInfo


class Properties:

    PROPERTY_LOOKUP_TABLE = '<L'
    PROPERTY_LOOKUP_TABLE_SZ = struct.calcsize(PROPERTY_LOOKUP_TABLE)

    PROPERTY = '<LL'
    PROPERTY_SZ = struct.calcsize(PROPERTY)

    nd_table_size: int
    props: list
    properties: OrderedDict

    @classmethod
    def from_data(cls, data: bytes, offset: int) -> Tuple['Properties', int]:
        self = cls()

        self.props = []
        self._qualifiers_done = False

        (
            property_count,
        ) = struct.unpack_from(cls.PROPERTY_LOOKUP_TABLE, data, offset=offset)
        offset += cls.PROPERTY_LOOKUP_TABLE_SZ

        for _ in range(property_count):
            # (property_name_ref, property_info_ref)
            prop = struct.unpack_from(cls.PROPERTY, data, offset=offset)
            offset += cls.PROPERTY_SZ
            self.props.append(prop)

        self.nd_table_size = (property_count - 1) // 4 + 1

        return self, offset

    def copy(self):
        cp = Properties()
        cp.props = self.props
        cp.nd_table_size = self.nd_table_size
        cp.properties = self.properties.copy()
        return cp

    def load(self, heap: bytes, nd_value_table: bytes):
        properties = []
        for (name_ref, info_ref) in self.props:
            if name_ref & 0x80000000:
                name = DICTIONARY_REFERENCE[name_ref & 0x7fffffff]
            else:
                name, _ = EncodedString.from_data(heap, name_ref)

            prop = PropertyInfo(heap, info_ref)

            properties.append((name, prop))

        self.properties = \
            OrderedDict(sorted(properties, key=lambda x: x[1].order))

    def set_prop_values(
            self,
            heap: bytes,
            nd_value_table: bytes,
            set_defaults: bool = False,
            ignore_missing: bool = False,
            ignore_defaults: bool = False):

        assert not set_defaults or not ignore_missing, \
            'set_default and ignore_missing cannot both be True'

        offset = self.nd_table_size

        for name, prop in tuple(self.properties.items()):
            # Let's get the default Values
            if prop.type & CimType.CIM_ARRAY_FLAG:
                fmt, size = '<L', 4
            else:
                fmt, size = CimType.get_cim_type_ref(prop.type)

            item_value, = \
                struct.unpack_from(fmt, nd_value_table, offset=offset)
            offset += size

            if item_value == 0xffffffff or \
                    item_value == 0x0 or \
                    (set_defaults and not prop.inherited_default):

                if ignore_missing:
                    del self.properties[name]
                    continue
                if set_defaults:
                    prop._set_type_default()
                elif ignore_defaults:
                    prop.value = None
                continue

            # TODO: what to do with prop.null_default ?
            prop._set_value(item_value, heap)

    def set_prop_defaults(self, nd_value_table: bytes):
        """NdTable

        see [MS-WMIO]: 2.2.26 NdTable and 2.2.27 NullAndDefaultFlag
        """
        nd_table = nd_value_table[:self.nd_table_size]

        nd_table = [
            (b >> shift) & 0b11
            for b in nd_table for shift in (0, 2, 4, 6)]

        for prop in self.properties.values():
            nd_entry = nd_table[prop.order]
            prop.null_default = bool(nd_entry & 1)
            prop.inherited_default = bool(nd_entry & 2)

    def set_qualifiers(self, heap: bytes):
        if self._qualifiers_done:
            return
        self._qualifiers_done = True
        for prop in self.properties.values():
            prop._set_qualifiers(heap)



import pickle
from logging import Logger
from typing import Optional


class KerberosCache:

    __slots__ = ('_file_path', '_tgs', '_tgt')

    def __init__(self, file_path: Optional[str] = None):
        """KerberosCacche: create an Instance per connection
        Argument `file_path` must be for example:

            /tmp/my-host.kerberos.cache.bin

        If None, memory is used (must keep the KerberosCache alive)
        """
        self._file_path = file_path
        self._tgt: Optional[tuple[bytes, bytes]] = None
        self._tgs: Optional[tuple[bytes, bytes]] = None

    def open(self, logger: Logger) -> tuple[
                Optional[tuple[bytes, bytes]],
                Optional[tuple[bytes, bytes]]]:
        if self._file_path is not None and \
                (self._tgs is None or self._tgt is None):
            try:
                with open(self._file_path, 'rb') as fp:
                    dump = pickle.load(fp)
                self._tgt = dump[0], dump[1]
                self._tgs = dump[2], dump[3]
            except Exception:
                logger.warning(f'Failed to load from: {self._file_path}')

        return self._tgt, self._tgs

    def write(self,
              tgt: tuple[bytes, bytes],
              tgs: tuple[bytes, bytes],
              logger: Logger):
        self._tgt = tgt
        self._tgs = tgs
        if self._file_path is not None:
            try:
                with open(self._file_path, 'wb') as fp:
                    pickle.dump(self._tgt + self._tgs, fp)
            except Exception as e:
                logger.warning(
                    f'Failed to write to: {self._file_path}: {e}')



import hashlib
import hmac
import random
import string
import socket
import struct
import datetime
from Crypto.Cipher import DES, ARC4
from Crypto.Hash import MD4
from typing import Tuple


KNOWN_DES_INPUT = b"THIS!MTB"


def pad4(size) -> int:
    return (4 - size % 4) % 4


def pad8(size) -> int:
    return (8 - size % 8) % 8


def pad(size, is_ndr64=False) -> int:
    return pad8(size) if is_ndr64 else pad4(size)


def gen_referent_id() -> int:
    return random.randint(1, 65535)


def gen_cid() -> bytes:
    top = (1 << 31)-1
    return struct.pack(
        "IIII",
        random.randrange(top),
        random.randrange(top),
        random.randrange(top),
        random.randrange(top))


def _des_expand_key(key, offset: int) -> bytes:
    # Expand the key from a 7-byte password key into a 8-byte DES key
    key = bytearray(key[offset:offset+7]).ljust(7, b'\x00')
    return bytes(bytearray((
        (((key[0] >> 1) & 0x7f) << 1)
        (((key[0] & 0x01) << 6 | ((key[1] >> 2) & 0x3f)) << 1)
        (((key[1] & 0x03) << 5 | ((key[2] >> 3) & 0x1f)) << 1)
        (((key[2] & 0x07) << 4 | ((key[3] >> 4) & 0x0f)) << 1)
        (((key[3] & 0x0f) << 3 | ((key[4] >> 5) & 0x07)) << 1)
        (((key[4] & 0x1f) << 2 | ((key[5] >> 6) & 0x03)) << 1)
        (((key[5] & 0x3f) << 1 | ((key[6] >> 7) & 0x01)) << 1)
        ((key[6] & 0x7f) << 1))))


def compute_lmhash(password: str):
    # This is done according to Samba's encryption
    # specification (docs/html/ENCRYPTION.html)
    password = password.upper().encode("latin-1")
    key = _des_expand_key(password, 0)
    lmhash = DES.new(key, DES.MODE_ECB).encrypt(KNOWN_DES_INPUT)
    key = _des_expand_key(password, 7)
    return lmhash + DES.new(key, DES.MODE_ECB).encrypt(KNOWN_DES_INPUT)


def compute_nthash(password: str):
    # This is done according to Samba's encryption
    # specification (docs/html/ENCRYPTION.html)
    hash = MD4.new()
    hash.update(password.encode('utf_16le'))
    return hash.digest()


def hmac_md5(key, data):
    hash = hmac.new(key, digestmod=hashlib.md5)
    hash.update(data)
    return hash.digest()


def ntowf_v2(user: str, password: str, domain: bytes):
    hash = hmac.new(compute_nthash(password), digestmod=hashlib.md5)
    hash.update(user.upper().encode('utf-16le') + domain)
    return hash.digest()


def get_rangom_bytes(length):
    return bytes(''.join(
        random.choice(string.digits+string.ascii_letters)
        for _ in range(length)), 'latin-1')


def hmac_md5(key: bytes, data: bytes):
    h = hmac.new(key, digestmod=hashlib.md5)
    h.update(data)
    return h.digest()


def encrypted_session_key(key_exchange_key, exported_session_key):
    return ARC4.new(key_exchange_key).encrypt(exported_session_key)


def read_string_bindings(data: bytes, offset: int) -> Tuple[list, int]:
    bindings = []
    data_len = len(data)

    while offset + 2 <= data_len:
        w_tower_id, = struct.unpack_from('<H', data, offset=offset)
        offset += 2

        if w_tower_id == 0:
            return bindings, offset

        current_pos = offset
        found = False
        while current_pos + 1 < data_len:
            if data[current_pos:current_pos+2] == b'\x00\x00':
                end = current_pos + 2
                found = True
                break
            current_pos += 2

        if not found:
            raise ValueError(
                'Malformed NDR: String binding terminator not found')

        raw = data[offset:end]
        try:
            decoded = raw.decode('utf-16le').rstrip('\x00')
            bindings.append((w_tower_id, decoded))
        except UnicodeDecodeError:
            raise ValueError(
                'Malformed NDR: Invalid UTF-16 encoding in string binding')

        offset = end

    return bindings, offset


def get_null(is_ndr64=False):
    return b'\x00\x00\x00\x00\x00\x00\x00\x00' \
        if is_ndr64 else b'\x00\x00\x00\x00'


def is_fqdn(target):
    try:
        socket.inet_aton(target)
    except Exception:
        try:
            target.index(':')
        except Exception:
            return True
    return False


def dt_fmt(dt: datetime.datetime) -> str:
    """Returns type datetime as a string according the Microsoft
    WMI Query Language (WQL) queries format.
    https://docs.microsoft.com/en-us/windows/win32/wmisdk/cim-datetime
    """
    minutes = dt.utcoffset().total_seconds() // 60
    return f"{dt.strftime('%Y-%m-%d %H:%M:%S')}{minutes:+04g}"



# MS/RPC Constants
MSRPC_REQUEST = 0x00
MSRPC_PING = 0x01
MSRPC_RESPONSE = 0x02
MSRPC_FAULT = 0x03
MSRPC_WORKING = 0x04
MSRPC_NOCALL = 0x05
MSRPC_REJECT = 0x06
MSRPC_ACK = 0x07
MSRPC_CL_CANCEL = 0x08
MSRPC_FACK = 0x09
MSRPC_CANCELACK = 0x0A
MSRPC_BIND = 0x0B
MSRPC_BINDACK = 0x0C
MSRPC_BINDNAK = 0x0D
MSRPC_ALTERCTX = 0x0E
MSRPC_ALTERCTX_R = 0x0F
MSRPC_AUTH3 = 0x10
MSRPC_SHUTDOWN = 0x11
MSRPC_CO_CANCEL = 0x12
MSRPC_ORPHANED = 0x13
MSRPC_RTS = 0x14

# MS/RPC Packet Flags
PFC_FIRST_FRAG = 0x01
PFC_LAST_FRAG = 0x02

# For PDU types bind, bind_ack, alter_context, and
# alter_context_resp, this flag MUST be interpreted as PFC_SUPPORT_HEADER_SIGN
MSRPC_SUPPORT_SIGN = 0x04

# For the
# remaining PDU types, this flag MUST be interpreted as PFC_PENDING_CANCEL.
MSRPC_PENDING_CANCEL = 0x04

PFC_RESERVED_1 = 0x08
PFC_CONC_MPX = 0x10
PFC_DID_NOT_EXECUTE = 0x20
PFC_MAYBE = 0x40
PFC_OBJECT_UUID = 0x80

# Auth Types - Security Providers
RPC_C_AUTHN_NONE = 0x00
RPC_C_AUTHN_GSS_NEGOTIATE = 0x09
RPC_C_AUTHN_WINNT = 0x0A
RPC_C_AUTHN_GSS_SCHANNEL = 0x0E
RPC_C_AUTHN_GSS_KERBEROS = 0x10
RPC_C_AUTHN_NETLOGON = 0x44
RPC_C_AUTHN_DEFAULT = 0xFF

# Auth Levels
RPC_C_AUTHN_LEVEL_NONE = 1
RPC_C_AUTHN_LEVEL_CONNECT = 2
RPC_C_AUTHN_LEVEL_CALL = 3
RPC_C_AUTHN_LEVEL_PKT = 4
RPC_C_AUTHN_LEVEL_PKT_INTEGRITY = 5
RPC_C_AUTHN_LEVEL_PKT_PRIVACY = 6

# Reasons for rejection of a context element,
# included in bind_ack result reason
rpc_provider_reason = {
    0: 'reason_not_specified',
    1: 'abstract_syntax_not_supported',
    2: 'proposed_transfer_syntaxes_not_supported',
    3: 'local_limit_exceeded',
    4: 'protocol_version_not_specified',
    8: 'authentication_type_not_recognized',
    9: 'invalid_checksum'
}

MSRPC_CONT_RESULT_ACCEPT = 0
MSRPC_CONT_RESULT_USER_REJECT = 1
MSRPC_CONT_RESULT_PROV_REJECT = 2

# Results of a presentation context negotiation
rpc_cont_def_result = {
    0: 'acceptance',
    1: 'user_rejection',
    2: 'provider_rejection'
}

# status codes, references:
# https://docs.microsoft.com/windows/desktop/Rpc/rpc-return-values
# https://msdn.microsoft.com/library/default.asp?url=/library/en-us/randz/
#    protocol/common_return_values.asp
# winerror.h
# https://www.opengroup.org/onlinepubs/9629399/apdxn.htm

ACCESS_DENIED = 0X00000005
AUTHENTICATION_TYPE_NOT_RECOGNIZED = 0X00000008
EPT_S_CANT_PERFORM_OP = 0X000006D8
RPC_S_INVALID_BOUND = 0X000006C6
RPC_S_CANNOT_SUPPORT = 0X000006E4
RPC_X_BAD_STUB_DATA = 0X000006F7
RPC_S_SEC_PKG_ERROR = 0x00000721
NCA_S_COMM_FAILURE = 0X1C010001
NCA_S_OP_RNG_ERROR = 0X1C010002
NCA_S_UNK_IF = 0X1C010003
NCA_S_WRONG_BOOT_TIME = 0X1C010006
NCA_S_YOU_CRASHED = 0X1C010009
NCA_S_PROTO_ERROR = 0X1C01000B
NCA_S_OUT_ARGS_TOO_BIG = 0X1C010013
NCA_S_SERVER_TOO_BUSY = 0X1C010014
NCA_S_FAULT_STRING_TOO_LONG = 0X1C010015
NCA_S_UNSUPPORTED_TYPE = 0X1C010017
NCA_S_FAULT_INT_DIV_BY_ZERO = 0X1C000001
NCA_S_FAULT_ADDR_ERROR = 0X1C000002
NCA_S_FAULT_FP_DIV_ZERO = 0X1C000003
NCA_S_FAULT_FP_UNDERFLOW = 0X1C000004
NCA_S_FAULT_FP_OVERFLOW = 0X1C000005
NCA_S_FAULT_INVALID_TAG = 0X1C000006
NCA_S_FAULT_INVALID_BOUND = 0X1C000007
NCA_S_RPC_VERSION_MISMATCH = 0X1C000008
NCA_S_UNSPEC_REJECT = 0X1C000009
NCA_S_BAD_ACTID = 0X1C00000A
NCA_S_WHO_ARE_YOU_FAILED = 0X1C00000B
NCA_S_MANAGER_NOT_ENTERED = 0X1C00000C
NCA_S_FAULT_CANCEL = 0X1C00000D
NCA_S_FAULT_ILL_INST = 0X1C00000E
NCA_S_FAULT_FP_ERROR = 0X1C00000F
NCA_S_FAULT_INT_OVERFLOW = 0X1C000010
NCA_S_FAULT_UNSPEC = 0X1C000012
NCA_S_FAULT_REMOTE_COMM_FAILURE = 0X1C000013
NCA_S_FAULT_PIPE_EMPTY = 0X1C000014
NCA_S_FAULT_PIPE_CLOSED = 0X1C000015
NCA_S_FAULT_PIPE_ORDER = 0X1C000016
NCA_S_FAULT_PIPE_DISCIPLINE = 0X1C000017
NCA_S_FAULT_PIPE_COMM_ERROR = 0X1C000018
NCA_S_FAULT_PIPE_MEMORY = 0X1C000019
NCA_S_FAULT_CONTEXT_MISMATCH = 0X1C00001A
NCA_S_FAULT_REMOTE_NO_MEMORY = 0X1C00001B
NCA_S_INVALID_PRES_CONTEXT_ID = 0X1C00001C
NCA_S_UNSUPPORTED_AUTHN_LEVEL = 0X1C00001D
NCA_S_INVALID_CHECKSUM = 0X1C00001F
NCA_S_INVALID_CRC = 0X1C000020
NCA_S_FAULT_USER_DEFINED = 0X1C000021
NCA_S_FAULT_TX_OPEN_FAILED = 0X1C000022
NCA_S_FAULT_CODESET_CONV_ERROR = 0X1C000023
NCA_S_FAULT_OBJECT_NOT_FOUND = 0X1C000024
NCA_S_FAULT_NO_CLIENT_STUB = 0X1C000025
RPC_S_MOD = 0X16C9A000
RPC_S_OP_RNG_ERROR = 0X16C9A001
RPC_S_CANT_CREATE_SOCKET = 0X16C9A002
RPC_S_CANT_BIND_SOCKET = 0X16C9A003
RPC_S_NOT_IN_CALL = 0X16C9A004
RPC_S_NO_PORT = 0X16C9A005
RPC_S_WRONG_BOOT_TIME = 0X16C9A006
RPC_S_TOO_MANY_SOCKETS = 0X16C9A007
RPC_S_ILLEGAL_REGISTER = 0X16C9A008
RPC_S_CANT_RECV = 0X16C9A009
RPC_S_BAD_PKT = 0X16C9A00A
RPC_S_UNBOUND_HANDLE = 0X16C9A00B
RPC_S_ADDR_IN_USE = 0X16C9A00C
RPC_S_IN_ARGS_TOO_BIG = 0X16C9A00D
RPC_S_STRING_TOO_LONG = 0X16C9A00E
RPC_S_TOO_MANY_OBJECTS = 0X16C9A00F
RPC_S_BINDING_HAS_NO_AUTH = 0X16C9A010
RPC_S_UNKNOWN_AUTHN_SERVICE = 0X16C9A011
RPC_S_NO_MEMORY = 0X16C9A012
RPC_S_CANT_NMALLOC = 0X16C9A013
RPC_S_CALL_FAULTED = 0X16C9A014
RPC_S_CALL_FAILED = 0X16C9A015
RPC_S_COMM_FAILURE = 0X16C9A016
RPC_S_RPCD_COMM_FAILURE = 0X16C9A017
RPC_S_ILLEGAL_FAMILY_REBIND = 0X16C9A018
RPC_S_INVALID_HANDLE = 0X16C9A019
RPC_S_CODING_ERROR = 0X16C9A01A
RPC_S_OBJECT_NOT_FOUND = 0X16C9A01B
RPC_S_CTHREAD_NOT_FOUND = 0X16C9A01C
RPC_S_INVALID_BINDING = 0X16C9A01D
RPC_S_ALREADY_REGISTERED = 0X16C9A01E
RPC_S_ENDPOINT_NOT_FOUND = 0X16C9A01F
RPC_S_INVALID_RPC_PROTSEQ = 0X16C9A020
RPC_S_DESC_NOT_REGISTERED = 0X16C9A021
RPC_S_ALREADY_LISTENING = 0X16C9A022
RPC_S_NO_PROTSEQS = 0X16C9A023
RPC_S_NO_PROTSEQS_REGISTERED = 0X16C9A024
RPC_S_NO_BINDINGS = 0X16C9A025
RPC_S_MAX_DESCS_EXCEEDED = 0X16C9A026
RPC_S_NO_INTERFACES = 0X16C9A027
RPC_S_INVALID_TIMEOUT = 0X16C9A028
RPC_S_CANT_INQ_SOCKET = 0X16C9A029
RPC_S_INVALID_NAF_ID = 0X16C9A02A
RPC_S_INVAL_NET_ADDR = 0X16C9A02B
RPC_S_UNKNOWN_IF = 0X16C9A02C
RPC_S_UNSUPPORTED_TYPE = 0X16C9A02D
RPC_S_INVALID_CALL_OPT = 0X16C9A02E
RPC_S_NO_FAULT = 0X16C9A02F
RPC_S_CANCEL_TIMEOUT = 0X16C9A030
RPC_S_CALL_CANCELLED = 0X16C9A031
RPC_S_INVALID_CALL_HANDLE = 0X16C9A032
RPC_S_CANNOT_ALLOC_ASSOC = 0X16C9A033
RPC_S_CANNOT_CONNECT = 0X16C9A034
RPC_S_CONNECTION_ABORTED = 0X16C9A035
RPC_S_CONNECTION_CLOSED = 0X16C9A036
RPC_S_CANNOT_ACCEPT = 0X16C9A037
RPC_S_ASSOC_GRP_NOT_FOUND = 0X16C9A038
RPC_S_STUB_INTERFACE_ERROR = 0X16C9A039
RPC_S_INVALID_OBJECT = 0X16C9A03A
RPC_S_INVALID_TYPE = 0X16C9A03B
RPC_S_INVALID_IF_OPNUM = 0X16C9A03C
RPC_S_DIFFERENT_SERVER_INSTANCE = 0X16C9A03D
RPC_S_PROTOCOL_ERROR = 0X16C9A03E
RPC_S_CANT_RECVMSG = 0X16C9A03F
RPC_S_INVALID_STRING_BINDING = 0X16C9A040
RPC_S_CONNECT_TIMED_OUT = 0X16C9A041
RPC_S_CONNECT_REJECTED = 0X16C9A042
RPC_S_NETWORK_UNREACHABLE = 0X16C9A043
RPC_S_CONNECT_NO_RESOURCES = 0X16C9A044
RPC_S_REM_NETWORK_SHUTDOWN = 0X16C9A045
RPC_S_TOO_MANY_REM_CONNECTS = 0X16C9A046
RPC_S_NO_REM_ENDPOINT = 0X16C9A047
RPC_S_REM_HOST_DOWN = 0X16C9A048
RPC_S_HOST_UNREACHABLE = 0X16C9A049
RPC_S_ACCESS_CONTROL_INFO_INV = 0X16C9A04A
RPC_S_LOC_CONNECT_ABORTED = 0X16C9A04B
RPC_S_CONNECT_CLOSED_BY_REM = 0X16C9A04C
RPC_S_REM_HOST_CRASHED = 0X16C9A04D
RPC_S_INVALID_ENDPOINT_FORMAT = 0X16C9A04E
RPC_S_UNKNOWN_STATUS_CODE = 0X16C9A04F
RPC_S_UNKNOWN_MGR_TYPE = 0X16C9A050
RPC_S_ASSOC_CREATION_FAILED = 0X16C9A051
RPC_S_ASSOC_GRP_MAX_EXCEEDED = 0X16C9A052
RPC_S_ASSOC_GRP_ALLOC_FAILED = 0X16C9A053
RPC_S_SM_INVALID_STATE = 0X16C9A054
RPC_S_ASSOC_REQ_REJECTED = 0X16C9A055
RPC_S_ASSOC_SHUTDOWN = 0X16C9A056
RPC_S_TSYNTAXES_UNSUPPORTED = 0X16C9A057
RPC_S_CONTEXT_ID_NOT_FOUND = 0X16C9A058
RPC_S_CANT_LISTEN_SOCKET = 0X16C9A059
RPC_S_NO_ADDRS = 0X16C9A05A
RPC_S_CANT_GETPEERNAME = 0X16C9A05B
RPC_S_CANT_GET_IF_ID = 0X16C9A05C
RPC_S_PROTSEQ_NOT_SUPPORTED = 0X16C9A05D
RPC_S_CALL_ORPHANED = 0X16C9A05E
RPC_S_WHO_ARE_YOU_FAILED = 0X16C9A05F
RPC_S_UNKNOWN_REJECT = 0X16C9A060
RPC_S_TYPE_ALREADY_REGISTERED = 0X16C9A061
RPC_S_STOP_LISTENING_DISABLED = 0X16C9A062
RPC_S_INVALID_ARG = 0X16C9A063
RPC_S_NOT_SUPPORTED = 0X16C9A064
RPC_S_WRONG_KIND_OF_BINDING = 0X16C9A065
RPC_S_AUTHN_AUTHZ_MISMATCH = 0X16C9A066
RPC_S_CALL_QUEUED = 0X16C9A067
RPC_S_CANNOT_SET_NODELAY = 0X16C9A068
RPC_S_NOT_RPC_TOWER = 0X16C9A069
RPC_S_INVALID_RPC_PROTID = 0X16C9A06A
RPC_S_INVALID_RPC_FLOOR = 0X16C9A06B
RPC_S_CALL_TIMEOUT = 0X16C9A06C
RPC_S_MGMT_OP_DISALLOWED = 0X16C9A06D
RPC_S_MANAGER_NOT_ENTERED = 0X16C9A06E
RPC_S_CALLS_TOO_LARGE_FOR_WK_EP = 0X16C9A06F
RPC_S_SERVER_TOO_BUSY = 0X16C9A070
RPC_S_PROT_VERSION_MISMATCH = 0X16C9A071
RPC_S_RPC_PROT_VERSION_MISMATCH = 0X16C9A072
RPC_S_SS_NO_IMPORT_CURSOR = 0X16C9A073
RPC_S_FAULT_ADDR_ERROR = 0X16C9A074
RPC_S_FAULT_CONTEXT_MISMATCH = 0X16C9A075
RPC_S_FAULT_FP_DIV_BY_ZERO = 0X16C9A076
RPC_S_FAULT_FP_ERROR = 0X16C9A077
RPC_S_FAULT_FP_OVERFLOW = 0X16C9A078
RPC_S_FAULT_FP_UNDERFLOW = 0X16C9A079
RPC_S_FAULT_ILL_INST = 0X16C9A07A
RPC_S_FAULT_INT_DIV_BY_ZERO = 0X16C9A07B
RPC_S_FAULT_INT_OVERFLOW = 0X16C9A07C
RPC_S_FAULT_INVALID_BOUND = 0X16C9A07D
RPC_S_FAULT_INVALID_TAG = 0X16C9A07E
RPC_S_FAULT_PIPE_CLOSED = 0X16C9A07F
RPC_S_FAULT_PIPE_COMM_ERROR = 0X16C9A080
RPC_S_FAULT_PIPE_DISCIPLINE = 0X16C9A081
RPC_S_FAULT_PIPE_EMPTY = 0X16C9A082
RPC_S_FAULT_PIPE_MEMORY = 0X16C9A083
RPC_S_FAULT_PIPE_ORDER = 0X16C9A084
RPC_S_FAULT_REMOTE_COMM_FAILURE = 0X16C9A085
RPC_S_FAULT_REMOTE_NO_MEMORY = 0X16C9A086
RPC_S_FAULT_UNSPEC = 0X16C9A087
UUID_S_BAD_VERSION = 0X16C9A088
UUID_S_SOCKET_FAILURE = 0X16C9A089
UUID_S_GETCONF_FAILURE = 0X16C9A08A
UUID_S_NO_ADDRESS = 0X16C9A08B
UUID_S_OVERRUN = 0X16C9A08C
UUID_S_INTERNAL_ERROR = 0X16C9A08D
UUID_S_CODING_ERROR = 0X16C9A08E
UUID_S_INVALID_STRING_UUID = 0X16C9A08F
UUID_S_NO_MEMORY = 0X16C9A090
RPC_S_NO_MORE_ENTRIES = 0X16C9A091
RPC_S_UNKNOWN_NS_ERROR = 0X16C9A092
RPC_S_NAME_SERVICE_UNAVAILABLE = 0X16C9A093
RPC_S_INCOMPLETE_NAME = 0X16C9A094
RPC_S_GROUP_NOT_FOUND = 0X16C9A095
RPC_S_INVALID_NAME_SYNTAX = 0X16C9A096
RPC_S_NO_MORE_MEMBERS = 0X16C9A097
RPC_S_NO_MORE_INTERFACES = 0X16C9A098
RPC_S_INVALID_NAME_SERVICE = 0X16C9A099
RPC_S_NO_NAME_MAPPING = 0X16C9A09A
RPC_S_PROFILE_NOT_FOUND = 0X16C9A09B
RPC_S_NOT_FOUND = 0X16C9A09C
RPC_S_NO_UPDATES = 0X16C9A09D
RPC_S_UPDATE_FAILED = 0X16C9A09E
RPC_S_NO_MATCH_EXPORTED = 0X16C9A09F
RPC_S_ENTRY_NOT_FOUND = 0X16C9A0A0
RPC_S_INVALID_INQUIRY_CONTEXT = 0X16C9A0A1
RPC_S_INTERFACE_NOT_FOUND = 0X16C9A0A2
RPC_S_GROUP_MEMBER_NOT_FOUND = 0X16C9A0A3
RPC_S_ENTRY_ALREADY_EXISTS = 0X16C9A0A4
RPC_S_NSINIT_FAILURE = 0X16C9A0A5
RPC_S_UNSUPPORTED_NAME_SYNTAX = 0X16C9A0A6
RPC_S_NO_MORE_ELEMENTS = 0X16C9A0A7
RPC_S_NO_NS_PERMISSION = 0X16C9A0A8
RPC_S_INVALID_INQUIRY_TYPE = 0X16C9A0A9
RPC_S_PROFILE_ELEMENT_NOT_FOUND = 0X16C9A0AA
RPC_S_PROFILE_ELEMENT_REPLACED = 0X16C9A0AB
RPC_S_IMPORT_ALREADY_DONE = 0X16C9A0AC
RPC_S_DATABASE_BUSY = 0X16C9A0AD
RPC_S_INVALID_IMPORT_CONTEXT = 0X16C9A0AE
RPC_S_UUID_SET_NOT_FOUND = 0X16C9A0AF
RPC_S_UUID_MEMBER_NOT_FOUND = 0X16C9A0B0
RPC_S_NO_INTERFACES_EXPORTED = 0X16C9A0B1
RPC_S_TOWER_SET_NOT_FOUND = 0X16C9A0B2
RPC_S_TOWER_MEMBER_NOT_FOUND = 0X16C9A0B3
RPC_S_OBJ_UUID_NOT_FOUND = 0X16C9A0B4
RPC_S_NO_MORE_BINDINGS = 0X16C9A0B5
RPC_S_INVALID_PRIORITY = 0X16C9A0B6
RPC_S_NOT_RPC_ENTRY = 0X16C9A0B7
RPC_S_INVALID_LOOKUP_CONTEXT = 0X16C9A0B8
RPC_S_BINDING_VECTOR_FULL = 0X16C9A0B9
RPC_S_CYCLE_DETECTED = 0X16C9A0BA
RPC_S_NOTHING_TO_EXPORT = 0X16C9A0BB
RPC_S_NOTHING_TO_UNEXPORT = 0X16C9A0BC
RPC_S_INVALID_VERS_OPTION = 0X16C9A0BD
RPC_S_NO_RPC_DATA = 0X16C9A0BE
RPC_S_MBR_PICKED = 0X16C9A0BF
RPC_S_NOT_ALL_OBJS_UNEXPORTED = 0X16C9A0C0
RPC_S_NO_ENTRY_NAME = 0X16C9A0C1
RPC_S_PRIORITY_GROUP_DONE = 0X16C9A0C2
RPC_S_PARTIAL_RESULTS = 0X16C9A0C3
RPC_S_NO_ENV_SETUP = 0X16C9A0C4
TWR_S_UNKNOWN_SA = 0X16C9A0C5
TWR_S_UNKNOWN_TOWER = 0X16C9A0C6
TWR_S_NOT_IMPLEMENTED = 0X16C9A0C7
RPC_S_MAX_CALLS_TOO_SMALL = 0X16C9A0C8
RPC_S_CTHREAD_CREATE_FAILED = 0X16C9A0C9
RPC_S_CTHREAD_POOL_EXISTS = 0X16C9A0CA
RPC_S_CTHREAD_NO_SUCH_POOL = 0X16C9A0CB
RPC_S_CTHREAD_INVOKE_DISABLED = 0X16C9A0CC
EPT_S_CANT_PERFORM_OP = 0X16C9A0CD
EPT_S_NO_MEMORY = 0X16C9A0CE
EPT_S_DATABASE_INVALID = 0X16C9A0CF
EPT_S_CANT_CREATE = 0X16C9A0D0
EPT_S_CANT_ACCESS = 0X16C9A0D1
EPT_S_DATABASE_ALREADY_OPEN = 0X16C9A0D2
EPT_S_INVALID_ENTRY = 0X16C9A0D3
EPT_S_UPDATE_FAILED = 0X16C9A0D4
EPT_S_INVALID_CONTEXT = 0X16C9A0D5
EPT_S_NOT_REGISTERED = 0X16C9A0D6
EPT_S_SERVER_UNAVAILABLE = 0X16C9A0D7
RPC_S_UNDERSPECIFIED_NAME = 0X16C9A0D8
RPC_S_INVALID_NS_HANDLE = 0X16C9A0D9
RPC_S_UNKNOWN_ERROR = 0X16C9A0DA
RPC_S_SS_CHAR_TRANS_OPEN_FAIL = 0X16C9A0DB
RPC_S_SS_CHAR_TRANS_SHORT_FILE = 0X16C9A0DC
RPC_S_SS_CONTEXT_DAMAGED = 0X16C9A0DD
RPC_S_SS_IN_NULL_CONTEXT = 0X16C9A0DE
RPC_S_SOCKET_FAILURE = 0X16C9A0DF
RPC_S_UNSUPPORTED_PROTECT_LEVEL = 0X16C9A0E0
RPC_S_INVALID_CHECKSUM = 0X16C9A0E1
RPC_S_INVALID_CREDENTIALS = 0X16C9A0E2
RPC_S_CREDENTIALS_TOO_LARGE = 0X16C9A0E3
RPC_S_CALL_ID_NOT_FOUND = 0X16C9A0E4
RPC_S_KEY_ID_NOT_FOUND = 0X16C9A0E5
RPC_S_AUTH_BAD_INTEGRITY = 0X16C9A0E6
RPC_S_AUTH_TKT_EXPIRED = 0X16C9A0E7
RPC_S_AUTH_TKT_NYV = 0X16C9A0E8
RPC_S_AUTH_REPEAT = 0X16C9A0E9
RPC_S_AUTH_NOT_US = 0X16C9A0EA
RPC_S_AUTH_BADMATCH = 0X16C9A0EB
RPC_S_AUTH_SKEW = 0X16C9A0EC
RPC_S_AUTH_BADADDR = 0X16C9A0ED
RPC_S_AUTH_BADVERSION = 0X16C9A0EE
RPC_S_AUTH_MSG_TYPE = 0X16C9A0EF
RPC_S_AUTH_MODIFIED = 0X16C9A0F0
RPC_S_AUTH_BADORDER = 0X16C9A0F1
RPC_S_AUTH_BADKEYVER = 0X16C9A0F2
RPC_S_AUTH_NOKEY = 0X16C9A0F3
RPC_S_AUTH_MUT_FAIL = 0X16C9A0F4
RPC_S_AUTH_BADDIRECTION = 0X16C9A0F5
RPC_S_AUTH_METHOD = 0X16C9A0F6
RPC_S_AUTH_BADSEQ = 0X16C9A0F7
RPC_S_AUTH_INAPP_CKSUM = 0X16C9A0F8
RPC_S_AUTH_FIELD_TOOLONG = 0X16C9A0F9
RPC_S_INVALID_CRC = 0X16C9A0FA
RPC_S_BINDING_INCOMPLETE = 0X16C9A0FB
RPC_S_KEY_FUNC_NOT_ALLOWED = 0X16C9A0FC
RPC_S_UNKNOWN_STUB_RTL_IF_VERS = 0X16C9A0FD
RPC_S_UNKNOWN_IFSPEC_VERS = 0X16C9A0FE
RPC_S_PROTO_UNSUPP_BY_AUTH = 0X16C9A0FF
RPC_S_AUTHN_CHALLENGE_MALFORMED = 0X16C9A100
RPC_S_PROTECT_LEVEL_MISMATCH = 0X16C9A101
RPC_S_NO_MEPV = 0X16C9A102
RPC_S_STUB_PROTOCOL_ERROR = 0X16C9A103
RPC_S_CLASS_VERSION_MISMATCH = 0X16C9A104
RPC_S_HELPER_NOT_RUNNING = 0X16C9A105
RPC_S_HELPER_SHORT_READ = 0X16C9A106
RPC_S_HELPER_CATATONIC = 0X16C9A107
RPC_S_HELPER_ABORTED = 0X16C9A108
RPC_S_NOT_IN_KERNEL = 0X16C9A109
RPC_S_HELPER_WRONG_USER = 0X16C9A10A
RPC_S_HELPER_OVERFLOW = 0X16C9A10B
RPC_S_DG_NEED_WAY_AUTH = 0X16C9A10C
RPC_S_UNSUPPORTED_AUTH_SUBTYPE = 0X16C9A10D
RPC_S_WRONG_PICKLE_TYPE = 0X16C9A10E
RPC_S_NOT_LISTENING = 0X16C9A10F
RPC_S_SS_BAD_BUFFER = 0X16C9A110
RPC_S_SS_BAD_ES_ACTION = 0X16C9A111
RPC_S_SS_WRONG_ES_VERSION = 0X16C9A112
RPC_S_FAULT_USER_DEFINED = 0X16C9A113
RPC_S_SS_INCOMPATIBLE_CODESETS = 0X16C9A114
RPC_S_TX_NOT_IN_TRANSACTION = 0X16C9A115
RPC_S_TX_OPEN_FAILED = 0X16C9A116
RPC_S_PARTIAL_CREDENTIALS = 0X16C9A117
RPC_S_SS_INVALID_CODESET_TAG = 0X16C9A118
RPC_S_MGMT_BAD_TYPE = 0X16C9A119
RPC_S_SS_INVALID_CHAR_INPUT = 0X16C9A11A
RPC_S_SS_SHORT_CONV_BUFFER = 0X16C9A11B
RPC_S_SS_ICONV_ERROR = 0X16C9A11C
RPC_S_SS_NO_COMPAT_CODESET = 0X16C9A11D
RPC_S_SS_NO_COMPAT_CHARSETS = 0X16C9A11E
DCE_CS_C_OK = 0X16C9A11F
DCE_CS_C_UNKNOWN = 0X16C9A120
DCE_CS_C_NOTFOUND = 0X16C9A121
DCE_CS_C_CANNOT_OPEN_FILE = 0X16C9A122
DCE_CS_C_CANNOT_READ_FILE = 0X16C9A123
DCE_CS_C_CANNOT_ALLOCATE_MEMORY = 0X16C9A124
RPC_S_SS_CLEANUP_FAILED = 0X16C9A125
RPC_SVC_DESC_GENERAL = 0X16C9A126
RPC_SVC_DESC_MUTEX = 0X16C9A127
RPC_SVC_DESC_XMIT = 0X16C9A128
RPC_SVC_DESC_RECV = 0X16C9A129
RPC_SVC_DESC_DG_STATE = 0X16C9A12A
RPC_SVC_DESC_CANCEL = 0X16C9A12B
RPC_SVC_DESC_ORPHAN = 0X16C9A12C
RPC_SVC_DESC_CN_STATE = 0X16C9A12D
RPC_SVC_DESC_CN_PKT = 0X16C9A12E
RPC_SVC_DESC_PKT_QUOTAS = 0X16C9A12F
RPC_SVC_DESC_AUTH = 0X16C9A130
RPC_SVC_DESC_SOURCE = 0X16C9A131
RPC_SVC_DESC_STATS = 0X16C9A132
RPC_SVC_DESC_MEM = 0X16C9A133
RPC_SVC_DESC_MEM_TYPE = 0X16C9A134
RPC_SVC_DESC_DG_PKTLOG = 0X16C9A135
RPC_SVC_DESC_THREAD_ID = 0X16C9A136
RPC_SVC_DESC_TIMESTAMP = 0X16C9A137
RPC_SVC_DESC_CN_ERRORS = 0X16C9A138
RPC_SVC_DESC_CONV_THREAD = 0X16C9A139
RPC_SVC_DESC_PID = 0X16C9A13A
RPC_SVC_DESC_ATFORK = 0X16C9A13B
RPC_SVC_DESC_CMA_THREAD = 0X16C9A13C
RPC_SVC_DESC_INHERIT = 0X16C9A13D
RPC_SVC_DESC_DG_SOCKETS = 0X16C9A13E
RPC_SVC_DESC_TIMER = 0X16C9A13F
RPC_SVC_DESC_THREADS = 0X16C9A140
RPC_SVC_DESC_SERVER_CALL = 0X16C9A141
RPC_SVC_DESC_NSI = 0X16C9A142
RPC_SVC_DESC_DG_PKT = 0X16C9A143
RPC_M_CN_ILL_STATE_TRANS_SA = 0X16C9A144
RPC_M_CN_ILL_STATE_TRANS_CA = 0X16C9A145
RPC_M_CN_ILL_STATE_TRANS_SG = 0X16C9A146
RPC_M_CN_ILL_STATE_TRANS_CG = 0X16C9A147
RPC_M_CN_ILL_STATE_TRANS_SR = 0X16C9A148
RPC_M_CN_ILL_STATE_TRANS_CR = 0X16C9A149
RPC_M_BAD_PKT_TYPE = 0X16C9A14A
RPC_M_PROT_MISMATCH = 0X16C9A14B
RPC_M_FRAG_TOOBIG = 0X16C9A14C
RPC_M_UNSUPP_STUB_RTL_IF = 0X16C9A14D
RPC_M_UNHANDLED_CALLSTATE = 0X16C9A14E
RPC_M_CALL_FAILED = 0X16C9A14F
RPC_M_CALL_FAILED_NO_STATUS = 0X16C9A150
RPC_M_CALL_FAILED_ERRNO = 0X16C9A151
RPC_M_CALL_FAILED_S = 0X16C9A152
RPC_M_CALL_FAILED_C = 0X16C9A153
RPC_M_ERRMSG_TOOBIG = 0X16C9A154
RPC_M_INVALID_SRCHATTR = 0X16C9A155
RPC_M_NTS_NOT_FOUND = 0X16C9A156
RPC_M_INVALID_ACCBYTCNT = 0X16C9A157
RPC_M_PRE_V2_IFSPEC = 0X16C9A158
RPC_M_UNK_IFSPEC = 0X16C9A159
RPC_M_RECVBUF_TOOSMALL = 0X16C9A15A
RPC_M_UNALIGN_AUTHTRL = 0X16C9A15B
RPC_M_UNEXPECTED_EXC = 0X16C9A15C
RPC_M_NO_STUB_DATA = 0X16C9A15D
RPC_M_EVENTLIST_FULL = 0X16C9A15E
RPC_M_UNK_SOCK_TYPE = 0X16C9A15F
RPC_M_UNIMP_CALL = 0X16C9A160
RPC_M_INVALID_SEQNUM = 0X16C9A161
RPC_M_CANT_CREATE_UUID = 0X16C9A162
RPC_M_PRE_V2_SS = 0X16C9A163
RPC_M_DGPKT_POOL_CORRUPT = 0X16C9A164
RPC_M_DGPKT_BAD_FREE = 0X16C9A165
RPC_M_LOOKASIDE_CORRUPT = 0X16C9A166
RPC_M_ALLOC_FAIL = 0X16C9A167
RPC_M_REALLOC_FAIL = 0X16C9A168
RPC_M_CANT_OPEN_FILE = 0X16C9A169
RPC_M_CANT_READ_ADDR = 0X16C9A16A
RPC_SVC_DESC_LIBIDL = 0X16C9A16B
RPC_M_CTXRUNDOWN_NOMEM = 0X16C9A16C
RPC_M_CTXRUNDOWN_EXC = 0X16C9A16D
RPC_S_FAULT_CODESET_CONV_ERROR = 0X16C9A16E
RPC_S_NO_CALL_ACTIVE = 0X16C9A16F
RPC_S_CANNOT_SUPPORT = 0X16C9A170
RPC_S_NO_CONTEXT_AVAILABLE = 0X16C9A171



import struct
from typing import Optional
from ..ndr.orpcthis import ORPCTHIS
from ..dtypes.lpwstr import LPWSTR
from ..ndr.remote_create_instance_response import RemoteCreateInstanceResponse
from ..uuid import uuid_to_bin
from ..tools import get_null

"""
ORPCTHIS

version
\x05\x00\x07\x00
flags
\x00\x00\x00\x00
reserved
\x00\x00\x00\x00
cid
\xa7\xfd\xa2\x1c\xd4~\x83k\t\x16\x87XQ\xa5vO
extensions
\x00\x00\x00\x00

referent id (33379)
c\x82\x00\x00


wsz_network_resource

\x0f\x00\x00\x00
\x00\x00\x00\x00
\x0f\x00\x00\x00

string
/\x00/\x00.\x00/\x00r\x00o\x00o\x00t\x00/\x00c\x00i\x00m\x00v\x002\x00\x00\x00

padding
\xbf\xbf

null
\x00\x00\x00\x00
l_flags
\x00\x00\x00\x00
null
\x00\x00\x00\x00
"""


class NTLMLogin:

    FMT = '<l'

    def __init__(
            self,
            wsz_network_resource: str,
            wsz_preferred_locale: Optional[str] = None,
            p_ctx: Optional[bytes] = None):

        self.wsz_network_resource = LPWSTR(wsz_network_resource)
        self.wsz_preferred_locale = LPWSTR(wsz_preferred_locale)
        self.l_flags = 0
        self.p_ctx = get_null() if p_ctx is None else p_ctx

    def get_data(self) -> bytes:

        wsz_network_resource = self.wsz_network_resource.get_data()
        wsz_preferred_locale = self.wsz_preferred_locale.get_data()

        data = \
            ORPCTHIS.get_data() + \
            wsz_network_resource + \
            wsz_preferred_locale + \
            struct.pack(self.FMT, self.l_flags) + \
            self.p_ctx

        return data



import struct
from typing import Union
from ..tools import gen_referent_id, get_null, pad


class LPWSTR:

    FMT32 = '<LLLL'
    FMT64 = '<QQQQ'

    FMT32_SZ = struct.calcsize(FMT32)

    FMT32_NULL = '<L'
    FMT64_NULL = '<Q'

    def __init__(self, inp: Union[str, bytes, None]):
        if isinstance(inp, bytes):
            assert 0
            pass
        else:
            self.string = inp  # None or string

    def get_data(self) -> bytes:
        if self.string is None:
            return get_null()

        data = self.string + '\x00'
        data = data.encode('utf-16le')

        padding = pad(len(data))
        n = len(data)

        maximum_count = actual_count = n // 2
        offset = 0
        referent_id = gen_referent_id()

        return struct.pack(
            self.FMT32,
            referent_id,
            maximum_count,
            offset,
            actual_count) + data + padding*b'\xbf'



from .dcom_const import IID_IRemoteSCMActivator
from .dcom_const import NDR_TransferSyntaxIdentifier
from .ntlm.auth_authenticate import NTLMAuthAuthenticate
from .ntlm.auth_negotiate import NTLMAuthNegotiate
from .rpc.auth_verifier_co import RpcAuthVerifierCo
from .rpc.bind import RpcBind
from .rpc.bind_ack import RpcBindAck
from .rpc.bind_nac import RpcBindNak
from .rpc.common import RpcCommon
from .rpc.alter_ctx_r import RpcAlterCtxR
from .rpc.const import MSRPC_ALTERCTX_R
from .rpc.const import MSRPC_AUTH3
from .rpc.const import MSRPC_BINDACK
from .rpc.const import MSRPC_BINDNAK
from .rpc.const import MSRPC_FAULT
from .rpc.const import MSRPC_RESPONSE
from .rpc.const import RPC_C_AUTHN_GSS_NEGOTIATE
from .rpc.const import RPC_C_AUTHN_WINNT
from .rpc.cont_elem import RpcContElem
from .rpc.fault import RpcFault
from .rpc.response import RpcResponse
from .tools import pad4


class Dcom:
    def __init__(self):
        self._max_xmit_frag = None  # set be negotiation (MSRPC_BINDACK)
        self._call_id = 1
        self._seq_num = 0

    def set_max_xmit_frag(self, max_xmit_frag: int):
        self._max_xmit_frag = max_xmit_frag

    def set_call_id(self, rpc_common: RpcCommon):
        rpc_common.call_id = self._call_id
        self._call_id += 1

    def get_new_call_id(self) -> int:
        call_id = self._call_id
        self._call_id += 1
        return call_id

    def get_seq_num(self):
        seq_num = self._seq_num
        self._seq_num += 1
        return seq_num

    def get_negotiate_ntlm_pkg(
            self,
            iid: bytes,
            ntlm_auth_negotiate: NTLMAuthNegotiate,
            auth_level: int,
            context_id: int):
        rpc_bind = RpcBind()

        rpc_cont_elem = RpcContElem(iid)
        rpc_cont_elem.add_transfer_syntax(NDR_TransferSyntaxIdentifier)
        rpc_bind.add_cont_elem(rpc_cont_elem)

        auth_pad_length = pad4(rpc_bind.freeze_context())

        auth_verifier, auth_length = RpcAuthVerifierCo.make(
            RPC_C_AUTHN_WINNT,
            auth_level,
            auth_pad_length,
            context_id,  # context id
            ntlm_auth_negotiate.get_data())

        rpc_bind.set_auth_verifier(auth_verifier, auth_length)
        self.set_call_id(rpc_bind)

        data = rpc_bind.get_data()
        return data

    def get_bind_kerberos_pkg(
            self,
            iid: bytes,
            ap_req_bytes: bytes,
            auth_level: int,
            context_id):
        rpc_bind = RpcBind()

        rpc_cont_elem = RpcContElem(iid)
        rpc_cont_elem.add_transfer_syntax(NDR_TransferSyntaxIdentifier)
        rpc_bind.add_cont_elem(rpc_cont_elem)

        auth_pad_length = pad4(rpc_bind.freeze_context())

        auth_verifier, auth_length = RpcAuthVerifierCo.make(
            RPC_C_AUTHN_GSS_NEGOTIATE,
            auth_level,         # RPC_C_AUTHN_LEVEL_PKT_PRIVACY (6)
            auth_pad_length,
            context_id,         # auth_context_id
            ap_req_bytes        # AP-REQ bytes
        )

        rpc_bind.set_auth_verifier(auth_verifier, auth_length)
        self.set_call_id(rpc_bind)

        data = rpc_bind.get_data()
        return data

    @staticmethod
    def get_authenticate_ntlm_pkg(
            ntlm_auth_authenticate: NTLMAuthAuthenticate,
            auth_level: int,
            context_id: int) -> bytes:

        rpc_common = RpcCommon()
        rpc_common.init(MSRPC_AUTH3)

        ntlm_auth_data = ntlm_auth_authenticate.get_data()

        auth_verifier, auth_length = RpcAuthVerifierCo.make(
            RPC_C_AUTHN_WINNT,
            auth_level,
            0,
            context_id,  # context id
            ntlm_auth_data)
        rpc_common.set_pdu_data(b'    ')
        rpc_common.set_auth_verifier(auth_verifier, auth_length)

        data = rpc_common.get_data()
        return data

    _DCOM_RPC_MAP = {
        MSRPC_BINDACK: RpcBindAck,
        MSRPC_BINDNAK: RpcBindNak,
        MSRPC_ALTERCTX_R: RpcAlterCtxR,
        MSRPC_RESPONSE: RpcResponse,
        MSRPC_FAULT: RpcFault,
    }



from __future__ import annotations
from typing import TYPE_CHECKING
import struct
from .common import RpcCommon
from ..tools import pad4
from .result import RpcResult
from .auth_verifier_co import RpcAuthVerifierCo
from .baseresp import RpcBaseResp

if TYPE_CHECKING:
    from ..dcom import Dcom

# /* common fields */
# u_int16 max_xmit_frag; /* 16:02 max transmit frag size */
# u_int16 max_recv_frag; /* 18:02 max receive frag size */
# u_int32 assoc_group_id; /* 20:04 returned assoc_group_id */
# port_any_t sec_addr; (length u16, char[]) /* 24:yy optional secondary address

# * for process incarnation; local port

# * part of address only */
# /* restore 4-octet alignment */
# u_int8 [size_is(align(4))] pad2;
# /* presentation context result list, including hints */
# p_result_list_t p_result_list; /* variable size */
# typedef struct {
# u_int8 n_results; /* count */

# u_int8 reserved; /* alignment pad, m.b.z. */

# u_int16 reserved2; /* alignment pad, m.b.z. */

# p_result_t [size_is(n_results)] p_results[];

# } p_result_list_t;
# /* optional authentication verifier */
# /* following fields present iff auth_length != 0 */
# auth_verifier_co_t auth_verifier; /* xx:yy */
# } rpcconn_bind_ack_hdr_t


class RpcBindAck(RpcBaseResp):

    __slots__ = ('sec_addr', 'auth', 'rpc_common')

    BIND_ACK_FMT = '<HHLH'
    BIND_ACK_FMT_SIZE = struct.calcsize(BIND_ACK_FMT)

    def __init__(self, dcom: Dcom, rpc_common: RpcCommon, data: bytes):
        offset = RpcCommon.COMMON_SIZE
        (
            max_xmit_frag,
            max_recv_frag,
            assoc_group_id,
            sec_addr_len
        ) = struct.unpack_from(
            self.BIND_ACK_FMT,
            data,
            offset)
        offset += self.BIND_ACK_FMT_SIZE
        self.sec_addr = data[offset:offset+sec_addr_len]
        offset += sec_addr_len
        offset += pad4(offset)

        n_results, _, _ = struct.unpack_from(
            RpcResult.RESULT_LIST_FMT,
            data,
            offset)

        offset += RpcResult.RESULT_LIST_FMT_SIZE

        while n_results:
            offset += 20  # skip transfer syntax
            n_results -= 1

        offset += pad4(offset)

        if rpc_common.auth_length:
            self.auth = RpcAuthVerifierCo(
                data,
                rpc_common.auth_length,
                offset=offset)
        else:
            self.auth = None

        self.rpc_common = rpc_common

        # set max ximit frag
        dcom.set_max_xmit_frag(max_xmit_frag)



import struct
from typing import Tuple
from ..cim_type import CimType


class Qualifier:

    FMT = '<LBL'
    FMT_SZ = struct.calcsize(FMT)

    @classmethod
    def from_data(cls, data: bytes, offset: int) -> Tuple['Qualifier', int]:
        self = cls()
        (
            self.name,
            self.flavor,
            self.type,
        ) = struct.unpack_from(cls.FMT, data, offset=offset)
        offset += cls.FMT_SZ

        key = self.type & (~CimType.CIM_ARRAY_FLAG)

        self.value, offset = CimType.get_value(key, data, offset=offset)

        return self, offset



from __future__ import annotations
import struct
from .const import WBEM_INFINITE
from .dcom_const import IID_IWbemFetchSmartEnum_bin
from .exceptions import WbemFalse, ServerNotOptimized
from .ndr.get_smart_enum_response import GetSmartEnumResponse
from .ndr.next_big_response import NextBigResponse
from .ndr.next_response import NextResponse
from .ndr.orpcthis import ORPCTHIS
from .ndr.query_response import QueryResponse
from .ndr.rem_query_interface_response import RemQueryInterfaceResponse
from .ndr.smart_response import SmartResponse
from .rpc.request import RpcRequest
from .rpc.response import RpcResponse
from .tools import get_null
from .uuid import uuid_to_bin


class QContext:
    def __init__(
            self,
            query,
            conn,
            flags,
            proto,
            timeout,
            skip_optimize):
        self._conn = conn
        self._flags = flags
        self._class_parts = {}
        self._proto = proto
        self._timeout = timeout
        self._skip_optimize = skip_optimize
        self._query = query
        self._interface = None
        self.next = self._next_slow

    def results(
            self,
            ignore_defaults: bool = False,
            ignore_missing: bool = False,
            load_qualifiers: bool = False) -> QContext:
        self._ignore_defaults = ignore_defaults
        self._ignore_missing = ignore_missing
        self._load_qualifiers = load_qualifiers
        return self

    async def __aenter__(self):
        await self.start()
        return self

    async def start(self):
        if self._conn._namespace != self._query.namespace:
            await self._conn.login_ntlm(
                self._proto, namespace=self._query.namespace)

        flags = struct.pack('<L', self._flags)
        pdu_data =\
            ORPCTHIS.get_data() +\
            self._query._language.get_data() +\
            self._query._query.get_data() +\
            flags +\
            get_null()

        # Both conn and quert are no longer required
        self._conn = None
        self._query = None

        request = RpcRequest(
            op_num=20, uuid_str=self._proto._interface.get_ipid())
        request.set_pdu_data(pdu_data)

        request_pkg = request.sign_data(self._proto)

        rpc_response: RpcResponse = \
            await self._proto.get_dcom_response(
                request_pkg,
                size=RpcResponse.SIZE,
                timeout=self._timeout)

        message = rpc_response.get_message(self._proto)

        self._interface = QueryResponse(message)

        if not self._skip_optimize:
            try:
                await self.optimize()
            except ServerNotOptimized:
                pass

    async def __aexit__(self, exc_type, exc, tb):
        await self.release()

    def __aiter__(self):
        return self

    async def __anext__(self):
        try:
            next_response = await self.next()
        except WbemFalse:
            raise StopAsyncIteration
        return next_response.get_properties(
            self._ignore_defaults,
            self._ignore_missing,
            self._load_qualifiers)

    @property
    def is_optimized(self) -> bool:
        return self.next is self._next_smart

    async def optimize(self):
        """RemQueryInterface."""
        ipid = uuid_to_bin(self._interface.get_ipid())

        c_refs, c_iids = 1, 1

        pdu_data =\
            ORPCTHIS.get_data(flags=0) +\
            ipid +\
            struct.pack('<LH', c_refs, c_iids) +\
            b'\xce\xce' +\
            struct.pack('<L', c_iids) +\
            IID_IWbemFetchSmartEnum_bin

        iremunknown = \
            self._proto._interface.scm_reply_info_data.ipid_rem_unknown

        request = RpcRequest(op_num=3, uuid_str=iremunknown)
        request.set_pdu_data(pdu_data)

        request_pkg = request.sign_data(self._proto)

        # optimizing shouldn't take long so we can use the default timeout
        rpc_response: RpcResponse = \
            await self._proto.get_dcom_response(request_pkg, RpcResponse.SIZE)

        message = rpc_response.get_message(self._proto)

        interface = RemQueryInterfaceResponse(message)

        pdu_data = ORPCTHIS.get_data(flags=0) + ipid
        request = RpcRequest(op_num=3, uuid_str=interface.get_ipid())
        request.set_pdu_data(pdu_data)

        request_pkg = request.sign_data(self._proto)

        # optimizing shouldn't take long so we can use the default timeout
        rpc_response: RpcResponse = \
            await self._proto.get_dcom_response(request_pkg, RpcResponse.SIZE)

        message = rpc_response.get_message(self._proto)

        # we asked for one public inferface, so we need to release this one
        await interface.rem_release(self._proto)

        interface = GetSmartEnumResponse(message)

        self._class_parts = {}
        self._interface = interface
        self.next = self._next_smart

    async def _next_smart(self, timeout: int = WBEM_INFINITE) -> NextResponse:
        """3.1.4.7.1 IWbemWCOSmartEnum::Next (Opnum 3)"""
        ucount = 1  # set the ucount fixed to one (1)
        param = struct.pack('<LL', timeout, ucount)
        pdu_data =\
            ORPCTHIS.get_data(flags=0) +\
            self._interface.proxy_guid +\
            param

        request = RpcRequest(op_num=3, uuid_str=self._interface.get_ipid())
        request.set_pdu_data(pdu_data)

        request_pkg = request.sign_data(self._proto)

        # Note: this timeout is not the time-out used in the package request
        #       the timout here refers to the protocol time-out of this lib.
        rpc_response: RpcResponse = \
            await self._proto.get_dcom_response(
                request_pkg,
                size=RpcResponse.SIZE,
                timeout=self._timeout)

        message = rpc_response.get_message(self._proto)
        return SmartResponse(message, self._class_parts)

    async def _next_slow(self, timeout: int = WBEM_INFINITE) -> NextResponse:
        """IEnumWbemClassObject_Next.

        lTimeout: MUST be the maximum amount of time, in milliseconds, that the
            IEnumWbemClassObject::Next method call allows to pass before it
            times out. If the constant WBEM_INFINITE (0xFFFFFFFF) is specified,
            the call MUST wait until one or more CIM objects are available. If
            the value 0x0 (WBEM_NO_WAIT) is specified, the call MUST return the
            available CIM objects, if any, at the time the call is made, and
            MUST NOT wait for any more objects.
        """
        ucount = 1  # set the ucount fixed to one (1)
        param = struct.pack('<LL', timeout, ucount)
        pdu_data =\
            ORPCTHIS.get_data(flags=0) +\
            param

        request = RpcRequest(op_num=4, uuid_str=self._interface.get_ipid())
        request.set_pdu_data(pdu_data)

        request_pkg = request.sign_data(self._proto)

        # Note: this timeout is not the time-out used in the package request
        #       the timout here refers to the protocol time-out of this lib.
        rpc_response: RpcResponse = \
            await self._proto.get_dcom_response(
                request_pkg,
                size=RpcResponse.SIZE,
                timeout=self._timeout)

        message = rpc_response.get_message(self._proto)
        return NextBigResponse(message)

    async def release(self):
        try:
            await self._interface.rem_release(self._proto)
        except Exception:
            pass
        self.next = None
        self._interface = None
        self._proto = None
        self._class_parts.clear()

    done = release   # alias



from datetime import datetime, timezone
from .asn1 import asn1_len, asn1_tag, asn1_gt, asn1_int, asn1_ostr
from .tools import encrypt_kerberos_rc4, decrypt_kerberos_rc4
from .const import OID_KERBEROS_V5, OID_MS_LEGACY_KRB, OID_SPNEGO


def wrap_gss_kerberos(ap_req_bytes):
    # MechTypes [0]
    mech_types = asn1_tag(0, b'\x30\x0b' + OID_MS_LEGACY_KRB)

    # MechToken [2] -> GSS-API Token wrapper (0x60)
    inner_token = OID_KERBEROS_V5 + b'\x01\x00' + ap_req_bytes
    gss_token = b'\x60' + asn1_len(inner_token) + inner_token

    # OctetString wrapper (0x04) MechToken [2]
    mech_token = asn1_tag(2, asn1_ostr(gss_token))

    # NegTokenInit Sequence [30] -> Wrapper [0]
    neg_token_body = mech_types + mech_token
    neg_token_seq = b'\x30' + asn1_len(neg_token_body) + neg_token_body
    neg_token_init = asn1_tag(0, neg_token_seq)

    # Application 0 (0x60) wrapper
    final_body = OID_SPNEGO + neg_token_init
    return b'\x60' + asn1_len(final_body) + final_body


def build_ap_req(username: str,
                 domain: str,
                 ticket: bytes,
                 service_session_key: bytes) -> bytes:
    now = datetime.now(timezone.utc)
    timestamp = now.strftime('%Y%m%d%H%M%SZ').encode('ascii')

    gss_data = b'\x10\x00\x00\x00' + b'\x00' * 16 + b'\x3e\x10\x00\x00'
    cksum_inner = (
        asn1_tag(0, asn1_int(32771)) +
        asn1_tag(1, asn1_ostr(gss_data))
    )
    cksum_asn1 = asn1_tag(3, b'\x30\x23' + cksum_inner)

    # CName [2]
    uname_bytes = b'\x1b' + asn1_len(username) + username.encode()
    cname_inner = (
        asn1_tag(0, asn1_int(1)) +
        asn1_tag(1, b'\x30' + asn1_len(uname_bytes) + uname_bytes)
    )
    cname_asn1 = \
        asn1_tag(2, b'\x30' + asn1_len(cname_inner) + cname_inner)

    # Realm [1]
    realm_asn1 = asn1_tag(1, b'\x1b' + asn1_len(domain) + domain.encode())

    # Authenticator Body (Plaintext)
    auth_body = (
        asn1_tag(0, asn1_int(5)) +
        realm_asn1 +
        cname_asn1 +
        cksum_asn1 +
        asn1_tag(4, asn1_int(now.microsecond)) +
        asn1_tag(5, asn1_gt(timestamp)) +
        asn1_tag(7, b'\x02\x01\x00')
    )

    # Encrypt authenticator
    auth_inner = b'\x30' + asn1_len(auth_body) + auth_body
    auth_asn1 = b'\x62' + asn1_len(auth_inner) + auth_inner
    enc_auth = encrypt_kerberos_rc4(service_session_key, 11, auth_asn1)

    etype_asn1 = asn1_tag(0, b'\x02\x01\x17')  # RC4-HMAC (23)
    cipher_asn1 = asn1_tag(2, asn1_ostr(enc_auth))
    enc_body = etype_asn1 + cipher_asn1
    enc_part = asn1_tag(4, b'\x30' + asn1_len(enc_body) + enc_body)

    ap_req_body = (
        asn1_tag(0, asn1_int(5)) +                      # pvno
        asn1_tag(1, asn1_int(14)) +                     # msg-type
        asn1_tag(2, b'\x03\x05\x00\x20\x00\x00\x00') +  # ap-options
        asn1_tag(3, ticket) +                           # ticket
        enc_part                                        # authenticator
    )

    inner_pdu = b'\x30' + asn1_len(ap_req_body) + ap_req_body
    return b'\x6e' + asn1_len(inner_pdu) + inner_pdu


def get_active_key(auth_bytes: bytes,
                   service_session_key: bytes) -> tuple[bytes, int]:
    active_key, seq_number = None, 0

    etype_idx = auth_bytes.find(b'\x02\x01\x17')
    if etype_idx >= 0:
        # Read new session key...
        idx_a2 = auth_bytes.find(b'\xa2', etype_idx)
        idx_04 = auth_bytes.find(b'\x04', idx_a2)

        pos = idx_04 + 1
        length_byte = auth_bytes[pos]
        pos += 1
        if length_byte & 0x80:
            n = length_byte & 0x7f
            length = int.from_bytes(auth_bytes[pos: pos + n], 'big')
            pos += n
        else:
            length = length_byte

        cipher_blob = auth_bytes[pos: pos + length]

        decrypted_raw = \
            decrypt_kerberos_rc4(service_session_key, 12, cipher_blob)
        asn1_data = decrypted_raw[8:]  # Skip 8 bytes confounder

        KEY_MARKER = b'\xa1\x12\x04\x10'
        key_idx = asn1_data.find(KEY_MARKER)
        if key_idx != -1:
            active_key = asn1_data[key_idx + 4: key_idx + 4 + 16]

        SEQ_MARKER = b'\xa3'
        seq_idx = asn1_data.find(SEQ_MARKER)
        if seq_idx != -1:
            int_tag_idx = asn1_data.find(b'\x02', seq_idx)
            if int_tag_idx != -1 and int_tag_idx < seq_idx + 4:
                int_len = asn1_data[int_tag_idx + 1]
                offset = int_tag_idx + 2
                seq_bytes = asn1_data[offset: offset + int_len]
                seq_number = int.from_bytes(seq_bytes, 'big')

    return active_key, seq_number



import struct
from ..cim_type import CimType
from .encoded_string import EncodedString
from ..dtypes.dt import dt_from_str


class EncodedValue:

    FMT = '<L'
    FMT_SZ = struct.calcsize(FMT)

    @classmethod
    def _get_array(cls, p_type: int, cim_type: int, entry: int, heap: bytes):

        num_items, = struct.unpack_from(cls.FMT, heap, offset=entry)
        num_items &= ~CimType.CIM_ARRAY_FLAG

        offset = entry + cls.FMT_SZ

        arr = []
        fmt, size = CimType.CIM_TYPES_REF[p_type]

        if cim_type == CimType.CIM_ARRAY_STRING:
            # We have an array of strings
            # First items are DWORDs with the string pointers
            # inside the heap. We don't need those ones
            offset += size * num_items
            # Let's now grab the strings
            for _ in range(num_items):
                item, offset = EncodedString.from_data(heap, offset)
                arr.append(item)
        elif cim_type == CimType.CIM_ARRAY_OBJECT:
            # Discard the pointers
            offset += size * num_items
            for item in range(num_items):
                assert 0

                (
                    encoding_len,
                ) = struct.unpack_from(cls.FMT, heap, offset=offset)
                offset += cls.FMT_SZ

                item, offset = ObjectBlock.from_data(heap, offset)
                # TODO: test parsing
                # msb = METHOD_SIGNATURE_BLOCK(heapData)
                # unit = ENCODING_UNIT()
                # unit['ObjectEncodingLength'] = msb['EncodingLength']
                # unit['ObjectBlock'] = msb['ObjectBlock']
                # array.append(unit)
                # heapData = heapData[msb['EncodingLength']+4:]
        else:
            for item in range(num_items):
                item, = struct.unpack_from(fmt, heap, offset=offset)
                arr.append(item)
                offset += size
        return arr

    @classmethod
    def get_value(cls, cim_type: int, entry: int, heap: bytes):
        from .object_block import ObjectBlock
        p_type = \
            cim_type & (~(CimType.CIM_ARRAY_FLAG | CimType.CIM_INHERITED_FLAG))

        if entry == 0xffffffff:
            return  # None

        if cim_type & CimType.CIM_ARRAY_FLAG:
            return cls._get_array(p_type, cim_type, entry, heap)

        if p_type == CimType.CIM_TYPE_BOOLEAN:
            return entry == 0xffff

        if p_type == CimType.CIM_TYPE_OBJECT:
            # If the value type is CIM-TYPE-OBJECT, the EncodedValue is a
            # HeapRef to the object encoded as an
            # ObjectEncodingLength (section 2.2.4) followed by an
            # ObjectBlock (section 2.2.5).

            # msb = METHOD_SIGNATURE_BLOCK(heapData)
            # unit = ENCODING_UNIT()
            # unit['ObjectEncodingLength'] = msb['EncodingLength']
            # unit['ObjectBlock'] = msb['ObjectBlock']
            # value = unit
            assert 0

        if p_type not in (
                CimType.CIM_TYPE_STRING,
                CimType.CIM_TYPE_DATETIME,
                CimType.CIM_TYPE_REFERENCE):
            return entry

        try:
            value, _ = EncodedString.from_data(heap, entry)
        except UnicodeDecodeError:
            raise  # TODO: dump heap data for debugging?

        if p_type == CimType.CIM_TYPE_DATETIME:
            return dt_from_str(value)  # Returns type datetime or timedelta

        return value



import asyncio
from typing import Optional, Callable
from Crypto.Cipher import ARC4
from .exceptions import AccessDenied, NoNewActiveKey, BindNak
from .kerberos.ap_req import build_ap_req, wrap_gss_kerberos, get_active_key
from .kerberos.cache import KerberosCache
from .kerberos.krb5_pdu import get_neg_token, build_alter_context
from .kerberos.tgs import get_tgs
from .kerberos.tgt import get_tgt
from .kerberos.tools import gss_unwrap_kerberos
from .kerberos.tools import sign_func_kerberos, seal_func_kerberos
from .ntlm.auth_authenticate import NTLMAuthAuthenticate
from .ntlm.auth_challange import NTLMAuthChallenge
from .ntlm.auth_negotiate import NTLMAuthNegotiate
from .ntlm.tools import seal_func, sign_func
from .protocol import Protocol
from .tools import get_rangom_bytes, encrypted_session_key
from .rpc.const import (
    RPC_C_AUTHN_LEVEL_CONNECT,
    RPC_C_AUTHN_LEVEL_PKT_INTEGRITY,
    RPC_C_AUTHN_LEVEL_PKT_PRIVACY,
    RPC_C_AUTHN_WINNT
)
from .ntlm.const import (
    NTLMSSP_NEGOTIATE_EXTENDED_SESSIONSECURITY,
    NTLMSSP_NEGOTIATE_KEY_EXCH,
)
from .ntlm.tools import (
    seal_func,
    seal_key,
    sign_func,
    sign_key,
)
from .ndr.remote_create_instance import RemoteCreateInstance
from .dcom_const import (
    CLSID_IWbemLevel1Login,
    IID_IRemoteSCMActivator,
    IID_IWbemLevel1Login,
    IID_IWbemServices,
)
from .rpc.bind_ack import RpcBindAck
from .rpc.request import RpcRequest
from .rpc.response import RpcResponse
from .ndr.remote_create_instance_response import RemoteCreateInstanceResponse
from .ndr.ntlm_login_response import NTLMLoginResponse
from .ntlm.login import NTLMLogin
from .logger import logger


class Connection:

    def __init__(
            self,
            host: str,
            username: str,
            password: str,
            domain: str = '',
            port: int = 135,
            kdc_host: Optional[str] = None,
            kdc_port: int = 88,
            kerberos_cache: Optional[KerberosCache] = None,
            loop: Optional[asyncio.AbstractEventLoop] = None):
        self._host = host
        self._port = port
        self._username = username
        self._password = password
        self._domain = domain
        self._kdc_host = kdc_host or self._host
        self._kdc_port = kdc_port
        self._loop = asyncio.get_event_loop() if loop is None else loop
        self._protocol: Optional[Protocol] = None
        self._timeout: int = 10
        self._namespace: Optional[str] = None
        self._kerberos_cache = kerberos_cache or KerberosCache()
        self._tgt, self._tgs = self._kerberos_cache.open(logger)

    async def connect(self, timeout: int = 10):
        conn = self._loop.create_connection(
            lambda: Protocol(loop=self._loop),
            host=self._host,
            port=self._port)

        _, self._protocol = await asyncio.wait_for(
            conn,
            timeout=timeout)
        self._timeout = timeout

    def close(self):
        if self._protocol:
            self._protocol.close()
        self._protocol = None

    def is_connected(self) -> bool:
        return bool(self._protocol)

    def connection_info(self) -> str:
        if not self.is_connected():
            return 'disconnected'
        return self._protocol.connection_info()

    async def _bind_ntlm(self, iid: bytes, proto: Protocol):
        ntlm_auth_negotiate = NTLMAuthNegotiate()

        ntlm_negotiate_pkg = \
            proto._dcom.get_negotiate_ntlm_pkg(
                iid,
                ntlm_auth_negotiate,
                proto._auth_level,
                proto._context_id)

        rpc_bind_ack: RpcBindAck = \
            await proto.get_dcom_response(
                ntlm_negotiate_pkg,
                timeout=self._timeout)

        proto._auth_type = rpc_bind_ack.auth.auth_type
        proto._auth_level = rpc_bind_ack.auth.auth_level

        ntlm_auth_challenge = NTLMAuthChallenge(rpc_bind_ack.auth.auth_value)
        proto._flags = flags = ntlm_auth_negotiate.get_negotiate_flags()

        ntlm_auth_authenticate = NTLMAuthAuthenticate(flags)

        session_base_key = ntlm_auth_authenticate.set_credentials(
            self._username,
            self._password,
            ntlm_auth_challenge.target_info,
            ntlm_auth_challenge.server_challenge,
            domain_name=self._domain)

        assert flags == 3767042613

        if flags & NTLMSSP_NEGOTIATE_KEY_EXCH:
            exported_session_key = get_rangom_bytes(16)

            encr_random_session_key = encrypted_session_key(
                session_base_key, exported_session_key)

            ntlm_auth_authenticate.set_encr_random_session_key(
                encr_random_session_key)
        else:
            exported_session_key = session_base_key

        ntlm_auth_authenticate.set_workstation_name(self._host)

        if proto._auth_level in (
                RPC_C_AUTHN_LEVEL_CONNECT,
                RPC_C_AUTHN_LEVEL_PKT_INTEGRITY,
                RPC_C_AUTHN_LEVEL_PKT_PRIVACY):
            if proto._auth_type == RPC_C_AUTHN_WINNT:
                if flags & NTLMSSP_NEGOTIATE_EXTENDED_SESSIONSECURITY:
                    client_signing_key = sign_key(
                        flags,
                        exported_session_key)
                    server_signing_key = sign_key(
                        flags,
                        exported_session_key,
                        client_mode=False)
                    client_sealing_key = seal_key(
                        flags,
                        exported_session_key)
                    server_sealing_key = seal_key(
                        flags,
                        exported_session_key,
                        client_mode=False)

                    client_sealing_handle = \
                        ARC4.new(client_sealing_key).encrypt
                    server_sealing_handle = \
                        ARC4.new(server_sealing_key).encrypt

                    proto._client_seal = seal_func(
                        client_signing_key,
                        client_sealing_handle)
                    proto._server_seal = seal_func(
                        server_signing_key,
                        server_sealing_handle)
                    proto._client_sign = sign_func(
                        client_signing_key,
                        client_sealing_handle)

        ntlm_auth_pkg = \
            proto._dcom.get_authenticate_ntlm_pkg(
                ntlm_auth_authenticate,
                proto._auth_level,
                proto._context_id)

        proto.write(ntlm_auth_pkg)

    async def _negotiate_kerberos(self):
        self._tgt = await get_tgt(self._username,
                                  self._password,
                                  self._domain,
                                  self._kdc_host,
                                  self._kdc_port)
        self._tgs = await get_tgs(self._username,
                                  self._domain,
                                  self._host,
                                  *self._tgt,
                                  self._kdc_host,
                                  self._kdc_port)

        self._kerberos_cache.write(self._tgt, self._tgs, logger)

    async def negotiate_kerberos(self) -> Protocol:
        if not self._domain:
            raise Exception('domain is required for Kerberos authentication')
        MAX_ATTEMPTS = 2
        attempt = 0
        has_keys = self._tgt is not None and self._tgt is not None
        while True:
            proto = self._protocol
            proto._auth_level = RPC_C_AUTHN_LEVEL_PKT_PRIVACY
            proto._context_id = 0x0001357f

            if not has_keys:
                logger.info('Start Kerberos negotion for TGT/TGS')
                await self._negotiate_kerberos()
                has_keys = True
            else:
                logger.info('Using Kerberos TGT/TGS from cache')
            try:
                await self._bind_kerberos(IID_IRemoteSCMActivator, proto)
                iface = await self._if_binding(proto,
                                               bind_func=self._bind_kerberos,
                                               m_auth_level=proto._auth_level)
            except (AccessDenied, NoNewActiveKey, BindNak) as e:
                msg = str(e) or type(e).__name__
                logger.warning(f'{msg} (attempt {attempt}/{MAX_ATTEMPTS})')
                if attempt <= MAX_ATTEMPTS:
                    if attempt:
                        has_keys = False  # attemp to get new keys
                    self.close()
                    await asyncio.sleep(0.5 * attempt)
                    attempt += 1
                    await self.connect(timeout=self._timeout)
                    continue
                raise
            else:
                break

        return iface

    async def _bind_kerberos(self, iid: bytes, proto: Protocol):
        ticket, service_session_key = self._tgs
        ap_req = build_ap_req(self._username,
                              self._domain,
                              ticket, service_session_key)
        auth_value = wrap_gss_kerberos(ap_req)
        bind_pkg = proto._dcom.get_bind_kerberos_pkg(
            iid,
            auth_value,
            proto._auth_level,
            proto._context_id,
        )
        rpc_bind_ack = await proto.get_dcom_response(bind_pkg)
        active_key, seq_number = get_active_key(rpc_bind_ack.auth.auth_value,
                                                service_session_key)
        if active_key is None:
            raise NoNewActiveKey()
        proto._auth_type = rpc_bind_ack.auth.auth_type
        proto._auth_level = rpc_bind_ack.auth.auth_level

        neg_token = get_neg_token(service_session_key, seq_number)
        alter_context_pkg = build_alter_context(
            iid,
            proto._dcom.get_new_call_id(),
            proto._auth_level,
            proto._context_id,
            neg_token,
        )
        _ = await proto.get_dcom_response(alter_context_pkg)

        if proto._auth_level >= RPC_C_AUTHN_LEVEL_PKT_INTEGRITY:
            proto._client_sign = sign_func_kerberos(active_key)
            proto._client_seal = seal_func_kerberos(active_key)
            proto._server_seal = gss_unwrap_kerberos(active_key)

    async def negotiate_ntlm(self) -> Protocol:
        proto = self._protocol
        proto._auth_level = RPC_C_AUTHN_LEVEL_PKT_PRIVACY
        proto._context_id = 4242

        await self._bind_ntlm(IID_IRemoteSCMActivator, proto)
        return await self._if_binding(proto, bind_func=self._bind_ntlm)

    async def login_ntlm(
            self,
            proto: Protocol,
            namespace: str = 'root/cimv2'):
        if not namespace.startswith('//'):
            namespace = '//./' + namespace
        ntlm_login = NTLMLogin(namespace)
        ntlm_login_pkg = ntlm_login.get_data()
        interface = self._protocol._interface

        request = RpcRequest(op_num=6, uuid_str=interface.get_ipid())

        request.set_pdu_data(ntlm_login_pkg)

        request_pkg = request.sign_data(proto)

        rpc_response: RpcResponse = \
            await proto.get_dcom_response(
                request_pkg,
                size=RpcResponse.SIZE,
                timeout=self._timeout)

        message = rpc_response.get_message(proto)

        ntlm_login_resp = NTLMLoginResponse(message)

        proto._interface = ntlm_login_resp
        proto._interface.scm_reply_info_data = interface.scm_reply_info_data
        self._iid = IID_IWbemServices
        self._namespace = namespace

    async def _if_binding(self,
                          proto: Protocol,
                          bind_func: Callable,
                          m_auth_level: int = RPC_C_AUTHN_LEVEL_PKT_INTEGRITY):
        remote_create_instance = RemoteCreateInstance(
            CLSID_IWbemLevel1Login,
            IID_IWbemLevel1Login)

        rci_pkg = remote_create_instance.get_data()

        request = RpcRequest(op_num=4)
        request.set_pdu_data(rci_pkg)

        request_pkg = request.seal_data(proto)

        rpc_response: RpcResponse = \
            await proto.get_dcom_response(
                request_pkg,
                size=RpcResponse.SIZE,
                timeout=self._timeout)

        message = rpc_response.get_message(proto)

        interface = RemoteCreateInstanceResponse(self._host, message)

        proto._interface = interface
        proto._iid = IID_IWbemLevel1Login

        host, port = interface.get_binding()

        try:
            conn = self._loop.create_connection(
                lambda: Protocol(loop=self._loop),
                host=self._host,
                port=port)

            _, proto = await asyncio.wait_for(
                conn,
                timeout=self._timeout)
        except Exception as e:
            logger.warning(
                f'failed to connect to {self._host} ({port}); '
                f'fallback to {host} ({port})...')

            conn = self._loop.create_connection(
                lambda: Protocol(loop=self._loop),
                host=host,
                port=port)

            _, proto = await asyncio.wait_for(
                conn,
                timeout=self._timeout)

        proto._auth_level = max(
            interface.scm_reply_info_data.authn_hint,
            m_auth_level)

        await bind_func(IID_IWbemLevel1Login, proto)
        return proto



import secrets
import struct
from Crypto.Hash import HMAC, MD5
from Crypto.Cipher import ARC4


GSS_WRAP_HEADER = b'\x60\x2b\x06\x09\x2a\x86\x48\x86\xf7\x12\x01\x02\x02'
GSS_WRAP_HEADER_LEN = len(GSS_WRAP_HEADER)  # 13


def gss_wrap_rc4(session_key: bytes,
                 data: bytes,
                 seq_num: int,
                 direction='init',
                 encrypt=True):
    pad = (8 - (len(data) % 8)) & 0x7
    pad_str = bytes([pad]) * pad
    data += pad_str

    seal_alg = b'\x10\x00' if encrypt else b'\xff\xff'
    token_header = b'\x02\x01\x11\x00' + seal_alg + b'\xff\xff'

    if direction == 'init':
        snd_seq = struct.pack('>L', seq_num) + b'\x00' * 4
    else:
        snd_seq = struct.pack('>L', seq_num) + b'\xff' * 4

    confounder = secrets.token_bytes(8)

    k_sign = HMAC.new(session_key, b'signaturekey\0', MD5).digest()

    md5_pre_hash = MD5.new(
        struct.pack('<L', 13) + token_header + confounder + data).digest()

    sgn_cksum = HMAC.new(k_sign, md5_pre_hash, MD5).digest()
    sgn_cksum_8 = sgn_cksum[:8]

    k_seq_base = HMAC.new(session_key, b'\x00\x00\x00\x00', MD5).digest()
    k_seq = HMAC.new(k_seq_base, sgn_cksum_8, MD5).digest()

    enc_snd_seq = ARC4.new(k_seq).encrypt(snd_seq)

    if encrypt:
        k_local = bytes([b ^ 0xF0 for b in session_key])

        k_crypt = HMAC.new(k_local, b'\x00\x00\x00\x00', MD5).digest()
        k_crypt = HMAC.new(k_crypt, struct.pack('>L', seq_num), MD5).digest()

        rc4 = ARC4.new(k_crypt)
        enc_confounder = rc4.encrypt(confounder)
        cipher_text = rc4.encrypt(data)
    else:
        enc_confounder = confounder
        cipher_text = data

    token_data = token_header + enc_snd_seq + sgn_cksum_8
    final_auth_data = GSS_WRAP_HEADER + token_data + enc_confounder

    return cipher_text, final_auth_data


def gss_unwrap_rc4(session_key: bytes,
                   cipher_text: bytes,
                   auth_data: bytes):
    token_bytes = auth_data[GSS_WRAP_HEADER_LEN:]
    sgn_cksum_8 = token_bytes[16:24]

    k_sign = HMAC.new(session_key, b'signaturekey\x00', MD5).digest()
    k_seq_base = HMAC.new(session_key, b'\x00\x00\x00\x00', MD5).digest()
    k_seq = HMAC.new(k_seq_base, sgn_cksum_8, MD5).digest()

    enc_snd_seq = token_bytes[8:16]
    snd_seq = ARC4.new(k_seq).decrypt(enc_snd_seq)

    k_local = bytes([b ^ 0xF0 for b in session_key])
    k_crypt_base = HMAC.new(k_local, b'\x00\x00\x00\x00', MD5).digest()
    k_crypt = HMAC.new(k_crypt_base, snd_seq[:4], MD5).digest()

    enc_confounder = auth_data[-8:]
    rc4 = ARC4.new(k_crypt)

    decrypted_blob = rc4.decrypt(enc_confounder + cipher_text)
    confounder = decrypted_blob[:8]
    data = decrypted_blob[8:]

    token_header = token_bytes[:8]
    md5_hash = MD5.new(
        struct.pack('<L', 13) + token_header + confounder + data).digest()
    expected_sgn_cksum = HMAC.new(k_sign, md5_hash, MD5).digest()

    if sgn_cksum_8 != expected_sgn_cksum[:8]:
        svr_seq = struct.unpack('>L', snd_seq[:4])[0]
        raise Exception(f"Integrity Check Failed! Server Seq was: {svr_seq}")

    return data



import struct
import random
from .common import NdrCommon
from ..tools import gen_referent_id
from ..dcom_const import CLSID_ScmRequestInfo


class ScmRequestInfoData(NdrCommon):

    CLSID = CLSID_ScmRequestInfo

    BUFFER32_FMT = '<lLLH'
    BUFFER64_FMT = '<lQLH'

    REFERENT32_FMT = '<LL'
    REFERENT64_FMT = '<LQ'

    @classmethod
    def get_data(cls) -> bytes:
        p_requested_prot_seqs = [7]

        dw_reserved = 0
        referent_id = gen_referent_id()
        client_imp_level = 0
        c_requested_prot_seqs = 1

        referents = b''.join([
            struct.pack('<H', x)
            for x in p_requested_prot_seqs])

        padding = b'\xaa\xaa'  # can be fiex as this is aways the same

        buffer = struct.pack(
            cls.BUFFER32_FMT,
            dw_reserved,            # <l
            referent_id,            # <L (32) / <Q (64)
            client_imp_level,       # <L
            c_requested_prot_seqs,  # <H
        ) + padding

        referent_id = gen_referent_id()

        buffer += struct.pack(
            cls.REFERENT32_FMT,
            referent_id,
            len(p_requested_prot_seqs),
        ) + referents

        private, padding = cls.get_private(buffer)

        return cls.get_common() + private + buffer + padding



class RC4():
    def __init__(self, key):
        bkey = bytearray(key)
        j = 0
        self.state = bytearray(range(256))
        for i in range(256):
            j = (j + self.state[i] + bkey[i % len(key)]) & 0xff
            self.state[i], self.state[j] = self.state[j], self.state[i]

    def encrypt(self, data):
        i = j = 0
        out = bytearray()
        for char in bytearray(data):
            i = (i+1) & 0xff
            j = (j+self.state[i]) & 0xff
            self.state[i], self.state[j] = self.state[j], self.state[i]
            out.append(
                char ^ self.state[(self.state[i] + self.state[j]) & 0xff])

        return bytes(out)

    def decrypt(self, data):
        return self.encrypt(data)



import struct
from .ndr_info_data import NdrInfoData
from ..tools import gen_referent_id
from .common import NdrCommon
from ..uuid import bin_to_str, CLSID_SZ


class ActivationBlob(NdrCommon):

    CLSID_NULL = b'\x00' * CLSID_SZ

    START_FMT = '<LL'
    START_SZ = struct.calcsize(START_FMT)

    BUF_FMT = '<LLLLL'
    BUF_SZ = struct.calcsize(BUF_FMT)

    CLSIDS32_FMT = '<LLLL'
    CLSIDS32_SZ = struct.calcsize(CLSIDS32_FMT)

    CLSIDS64_FMT = '<QQLL'  # not sure if the second is a referentId
    CLSIDS64_SZ = struct.calcsize(CLSIDS64_FMT)

    def __init__(self, dest_ctx: int = 2):
        self.dest_ctx = dest_ctx
        self.pclsid = []
        self.psizes = []
        self.properties = []

    @classmethod
    def from_data(cls, data: bytes):
        offset = 0
        self = cls()
        (
            total_size,
            dw_reserved,
        ) = struct.unpack_from(cls.START_FMT, data, offset=offset)
        offset += cls.START_SZ

        self.read_common(data, offset=offset)
        offset += self.COMMON_SZ

        self.read_private(data, offset=offset)
        offset += self.PRIVATE_SZ

        (
            self.total_size,
            self.header_size,
            dw_reserved,
            self.dest_ctx,
            c_ifs,
        ) = struct.unpack_from(cls.BUF_FMT, data, offset=offset)
        offset += cls.BUF_SZ

        self.clsid = bin_to_str(data, offset=offset)
        offset += CLSID_SZ

        (
            self.referent_id1,
            self.referent_id2,
            dw_reserved,
            c_clsids,
        ) = struct.unpack_from(cls.CLSIDS32_FMT, data, offset=offset)
        offset += cls.CLSIDS32_SZ

        for _ in range(c_clsids):
            clsid = bin_to_str(data, offset=offset)
            offset += CLSID_SZ
            self.pclsid.append(clsid)

        n_psizes, = struct.unpack_from('<L', data, offset=offset)
        offset += 4  # size(<L)

        for _ in range(n_psizes):
            size, = struct.unpack_from('<L', data, offset=offset)
            offset += 4  # size(<L)
            self.psizes.append(size)

        for size in self.psizes:
            self.properties.append(data[offset: offset+size])
            offset += size

        return self

    def add_info_data(self, info_data: NdrInfoData):
        data = info_data.get_data()
        self.pclsid.append(info_data.CLSID)
        self.psizes.append(len(data))
        self.properties.append(data)

    def get_data(self):
        dw_reserved = 0
        c_psizes = c_clsids = c_ifs = len(self.pclsid)

        referent_id1 = gen_referent_id()
        referent_id2 = gen_referent_id()

        psizes = \
            b''.join([struct.pack('<L', x) for x in [c_psizes] + self.psizes])
        clsids = b''.join(self.pclsid)
        properties = b''.join(self.properties)

        # TODO: should we include padding to the header size?
        # probably the total size should include padding before the properties?
        header_size = \
            self.COMMON_SZ + \
            self.PRIVATE_SZ + \
            self.BUF_SZ + \
            CLSID_SZ + \
            self.CLSIDS32_SZ + \
            len(clsids) + \
            len(psizes)
        total_size = header_size + len(properties)

        buffer = struct.pack(
            self.BUF_FMT,
            total_size,
            header_size,
            dw_reserved,
            self.dest_ctx,
            c_ifs,
        ) + self.CLSID_NULL + struct.pack(
            self.CLSIDS32_FMT,
            referent_id1,
            referent_id2,
            dw_reserved,
            c_clsids,
        ) + clsids + psizes

        private, padding = self.get_private(buffer)

        data = struct.pack(
            self.START_FMT,
            total_size,
            dw_reserved
        ) + self.get_common() + private + buffer + padding + properties

        return data



import struct
from ..exceptions import ServerNotOptimized
from .interface import NdrInterface
from .objref_std import ObjRefStd
from .orpcthat import ORPCTHAT


class RemQueryInterfaceResponse(NdrInterface):
    FMT1_32 = '<LlLL'
    FMT1_64 = '<QlLL'

    FMT1_32_SZ = struct.calcsize(FMT1_32)

    def __init__(self, data: bytes):
        self.orpcthat, offset = ORPCTHAT.from_data(data, offset=0)

        (
            self.referent_id,
            n_size,
            h_result,
            pad,
        ) = struct.unpack_from(self.FMT1_32, data, offset)
        offset += self.FMT1_32_SZ

        if h_result:
            raise ServerNotOptimized('Server is not optimized')

        self.objref, offset = ObjRefStd.from_data(data, offset)

        self.error_code, = struct.unpack_from('<L', data, offset)
        assert self.error_code == 0, f'error code: {self.error_code}'

    def get_ipid(self):
        return self.objref.ipid
